import logging
from collections.abc import Generator
from typing import Any

from datapizza.core.vectorstore import VectorConfig, Vectorstore
from datapizza.type import (
    Chunk,
    DenseEmbedding,
    Embedding,
    EmbeddingFormat,
    SparseEmbedding,
)
from pymilvus import (
    AsyncMilvusClient,
    CollectionSchema,
    DataType,
    FieldSchema,
    MilvusClient,
    MilvusException,
)
from pymilvus.milvus_client.index import IndexParams

log = logging.getLogger(__name__)


class MilvusVectorstore(Vectorstore):
    """
    Milvus Vectorstore
    """

    def __init__(
        self,
        # You can pass either `uri="http://localhost:19530"` | "./milvus.db" (Milvus Lite)
        # or host/port(+secure/user/password) via **connection_args for flexibility.
        uri: str | None = None,
        host: str | None = None,
        port: int | None = None,
        user: str | None = None,
        password: str | None = None,
        secure: bool | None = None,
        **connection_args: Any,
    ):
        self.conn_kwargs: dict[str, Any] = {}
        if uri:
            self.conn_kwargs["uri"] = uri
        if host:
            self.conn_kwargs["host"] = host
        if port:
            self.conn_kwargs["port"] = port
        if user:
            self.conn_kwargs["user"] = user
        if password:
            self.conn_kwargs["password"] = password
        if secure is not None:
            self.conn_kwargs["secure"] = secure

        # Allow extra MilvusClient kwargs (e.g. token for Zilliz)
        self.conn_kwargs.update(connection_args or {})
        self.client: MilvusClient
        self.a_client: AsyncMilvusClient
        self.batch_size: int = 100

    def get_client(self) -> MilvusClient:
        if not hasattr(self, "client"):
            self._init_client()
        return self.client

    def _get_a_client(self) -> AsyncMilvusClient:
        if not hasattr(self, "a_client"):
            self._init_a_client()
        return self.a_client

    def _init_client(self):
        self.client = MilvusClient(**self.conn_kwargs)

    def _init_a_client(self):
        self.a_client = AsyncMilvusClient(**self.conn_kwargs)

    @staticmethod
    def _chunk_to_row(chunk: Chunk) -> dict[str, Any]:
        def _sparse_to_dict(se: SparseEmbedding) -> dict[int, float]:
            return {
                int(i): float(v) for i, v in zip(se.indices, se.values, strict=False)
            }

        if not chunk.embeddings:
            raise ValueError("Chunk must have an embedding")

        row: dict[str, Any] = {
            "id": chunk.id,
            "text": chunk.text or "",
        }
        if chunk.metadata:
            row.update(chunk.metadata)

        if len(chunk.embeddings) == 1:
            e = chunk.embeddings[0]
            if isinstance(e, DenseEmbedding):
                row[e.name] = e.vector
            elif isinstance(e, SparseEmbedding):
                row[e.name] = _sparse_to_dict(e)
            else:
                raise ValueError(f"Unsupported embedding type: {type(e)}")
        else:
            for e in chunk.embeddings:
                if isinstance(e, DenseEmbedding):
                    fname = e.name
                    row[fname] = e.vector
                elif isinstance(e, SparseEmbedding):
                    fname = e.name
                    row[fname] = _sparse_to_dict(e)
                else:
                    raise ValueError(f"Unsupported embedding type: {type(e)}")
        return row

    @staticmethod
    def _entity_to_chunk(entity: dict[str, Any]) -> Chunk:
        """
        Convert a Milvus entity dict into a Chunk.
        - Dense vectors: list[float] -> DenseEmbedding(name=<field>)
        - Sparse vectors: dict[index->value] -> SparseEmbedding(name=<field>)
        - Everything else (except id/text) -> metadata
        """

        embeddings: list[Embedding] = []
        metadata: dict[str, Any] = {}

        for key, val in entity.items():
            if key in {"id", "text"}:
                continue

            if isinstance(val, list) and all(isinstance(x, float) for x in val):
                embeddings.append(DenseEmbedding(name=key, vector=list(val)))
                continue

            if isinstance(val, dict) and val:
                try:
                    items = sorted(
                        ((int(k), float(v)) for k, v in val.items()), key=lambda t: t[0]
                    )
                    indices = [i for i, _ in items]
                    values = [v for _, v in items]
                    embeddings.append(
                        SparseEmbedding(name=key, indices=indices, values=values)
                    )
                    continue
                except Exception:
                    # fall through to metadata if it isn't a proper sparse map
                    pass

            # Non-vector -> metadata
            metadata[key] = val

        return Chunk(
            id=entity["id"],
            text=entity["text"],
            embeddings=embeddings,
            metadata=metadata,
        )

    @staticmethod
    def _metric_from_config(cfg: VectorConfig) -> str:
        """Map your VectorConfig.distance to Milvus metric string."""
        v = (getattr(cfg.distance, "value", str(cfg.distance)) or "").upper()
        if "COS" in v:
            return "COSINE"
        return "L2"

    @staticmethod
    def _sparse_embedding_to_milvus_format(
        sparse_emb: SparseEmbedding,
    ) -> dict[int, float]:
        """
        Convert a SparseEmbedding instance to a Milvus-supported sparse vector format (dict).

        Milvus accepts: {dimension_index: value, ...}
        """
        return dict(zip(sparse_emb.indices, sparse_emb.values, strict=False))

    def add(
        self,
        chunk: Chunk | list[Chunk],
        collection_name: str | None = None,
        batch_size: int = 100,
    ):
        client = self.get_client()

        chunks = chunk if isinstance(chunk, list) else [chunk]
        total = len(chunks)

        try:
            for i in range(0, total, batch_size):
                batch = chunks[i : i + batch_size]
                rows = [self._chunk_to_row(c) for c in batch]
                client.insert(collection_name=collection_name, data=rows)

        except MilvusException as e:
            log.error(f"Failed to batch insert into '{collection_name}': {e!s}")
            raise

    async def a_add(
        self,
        chunk: Chunk | list[Chunk],
        collection_name: str | None = None,
        batch_size: int = 100,
    ):
        client = self._get_a_client()

        chunks = chunk if isinstance(chunk, list) else [chunk]
        total = len(chunks)

        try:
            for i in range(0, total, batch_size):
                batch = chunks[i : i + batch_size]
                rows = [self._chunk_to_row(c) for c in batch]
                await client.insert(collection_name=collection_name, data=rows)

        except MilvusException as e:
            log.error(f"Failed to batch insert into '{collection_name}': {e!s}")
            raise

    def retrieve(
        self,
        collection_name: str,
        ids: list[str],
        **kwargs,
    ) -> list[Chunk]:
        """
        Retrieve chunks from a collection by their IDs.

        Args:
            collection_name (str): Name of the collection to query.
            ids (list[str]): IDs to retrieve.
            **kwargs: Extra arguments to Milvus client query().

        Returns:
            list[Chunk]: Retrieved chunks.
        """

        client = self.get_client()
        try:
            res = client.query(
                collection_name=collection_name,
                ids=ids,
                **kwargs,
            )
        except Exception as e:
            log.error(f"Failed to retrieve from collection '{collection_name}': {e!s}")
            raise

        return [self._entity_to_chunk(r) for r in (res or [])]

    def remove(self, collection_name: str, ids: list[str], **kwargs):
        client = self.get_client()
        client.delete(collection_name=collection_name, ids=ids, **kwargs)

    def update(self, collection_name: str, chunk: Chunk | list[Chunk], **kwargs):
        """
        Upsert one or more Chunk objects into a Milvus collection.

        In Milvus, an upsert operation combines both insert and update behavior:
          - If an entity with the same primary key already exists in the collection,
            it will be overwritten with the new data.
          - If the primary key does not exist, a new entity will be inserted.

        Args:
            collection_name (str): Name of the Milvus collection.
            chunk (Chunk | list[Chunk]): One or more Chunk objects to upsert.
            **kwargs: Additional parameters passed to the Milvus client.

        """
        client = self.get_client()
        chunks = [chunk] if isinstance(chunk, Chunk) else chunk
        data = [self._chunk_to_row(c) for c in chunks]
        client.upsert(collection_name=collection_name, data=data, **kwargs)
        client.flush(collection_name=collection_name)

    @staticmethod
    def _merge_output_fields(user_fields: list[str] | None) -> list[str]:
        required = ["id", "text"]
        if user_fields is None:
            return required[:]
        if not isinstance(user_fields, list) or not all(
            isinstance(f, str) for f in user_fields
        ):
            raise TypeError("output_fields must be a list[str]")
        # ensure required fields are included
        return required[:] + [f for f in user_fields if f not in required]

    @staticmethod
    def _is_multi_query(vec) -> bool:
        return (
            isinstance(vec, list)
            and vec
            and isinstance(vec[0], (list, DenseEmbedding, SparseEmbedding))
        )

    def _normalize_query(
        self,
        v: list[float] | DenseEmbedding | SparseEmbedding,
        vector_name: str | None,
    ) -> tuple[list[float] | dict[int, float], str | None]:
        if isinstance(v, (DenseEmbedding, SparseEmbedding)) and not vector_name:
            vector_name = v.name
        if isinstance(v, SparseEmbedding):
            return self._sparse_embedding_to_milvus_format(v), vector_name
        if isinstance(v, DenseEmbedding):
            return v.vector, vector_name
        if not isinstance(v, list) or (v and not isinstance(v[0], (int, float))):
            raise TypeError(
                "query_vector must be list[float], DenseEmbedding, or SparseEmbedding"
            )
        return v, vector_name

    def _prepare_search_args(
        self,
        *,
        query_vector: list[float] | DenseEmbedding | SparseEmbedding,
        vector_name: str | None,
        k: int,
        kwargs: dict,
    ) -> dict:
        # guard against accidental multi-vector input
        if self._is_multi_query(query_vector):
            raise TypeError(
                "Single-vector search only: got a list of vectors. Pass exactly one vector."
            )

        # normalize
        vector, vector_name = self._normalize_query(query_vector, vector_name)
        if not vector_name:
            raise ValueError(
                "vector_name must be provided (or embedded in the Embedding.name)."
            )

        # fields
        user_fields = kwargs.pop("output_fields", None)
        # ensure at least id and text are included
        output_fields = self._merge_output_fields(user_fields)

        # final param bundle for Milvus .search()
        return {
            "data": [vector],
            "anns_field": vector_name,
            "limit": k,
            "output_fields": output_fields,
            **kwargs,
        }

    def _chunks_from_results(self, res) -> list[Chunk]:
        hits = res[0] if res else []
        out: list[Chunk] = []
        for h in hits:
            entity = h.get("entity", None)
            if entity is None:
                entity = {k: v for k, v in h.items() if k not in {"score", "distance"}}
            out.append(self._entity_to_chunk(entity))
        return out

    def search(
        self,
        collection_name: str,
        query_vector: list[float] | DenseEmbedding | SparseEmbedding,
        k: int = 10,
        vector_name: str | None = None,
        **kwargs,
    ) -> list[Chunk]:
        """
        Perform a single-vector similarity search on a Milvus collection.

        Args:
            collection_name (str): Name of the Milvus collection to search.
            query_vector (list[float] | DenseEmbedding | SparseEmbedding):
                The query vector or embedding to search for.
            k (int, optional): Number of nearest results to return. Defaults to 10.
            vector_name (str, optional): Name of the vector field to search on.
                If not provided, inferred from the embedding (if applicable).
            **kwargs: Additional parameters passed directly to the Milvus client's
                `search()` method (e.g., filter expressions, consistency level).

        Returns:
            list[Chunk]: A list of retrieved `Chunk` objects representing the top
            `k` most similar results.
        """
        client = self.get_client()
        params = self._prepare_search_args(
            query_vector=query_vector,
            vector_name=vector_name,
            k=k,
            kwargs=kwargs,
        )
        res = client.search(collection_name=collection_name, **params)
        return self._chunks_from_results(res)

    # ---------- async API ----------

    async def a_search(
        self,
        collection_name: str,
        query_vector: list[float] | DenseEmbedding | SparseEmbedding,
        k: int = 10,
        vector_name: str | None = None,
        **kwargs,
    ) -> list[Chunk]:
        """
        Perform an asynchronous single-vector similarity search on a Milvus collection.

        Args:
            collection_name (str): Name of the Milvus collection to search.
            query_vector (list[float] | DenseEmbedding | SparseEmbedding):
                The query vector or embedding to search for.
            k (int, optional): Number of nearest results to return. Defaults to 10.
            vector_name (str, optional): Name of the vector field to search on.
                If not provided, inferred from the embedding (if applicable).
            **kwargs: Additional parameters passed directly to the Milvus client's
                `search()` method (e.g., filter expressions, consistency level).

        Returns:
            list[Chunk]: A list of retrieved `Chunk` objects representing the top
            `k` most similar results.
        """
        client = self._get_a_client()
        params = self._prepare_search_args(
            query_vector=query_vector,
            vector_name=vector_name,
            k=k,
            kwargs=kwargs,
        )
        res = await client.search(collection_name=collection_name, **params)
        return self._chunks_from_results(res)

    def get_collections(self):
        client = self.get_client()
        return client.list_collections()

    def create_collection(
        self,
        collection_name: str,
        vector_config: list[VectorConfig],
        index_params: IndexParams = None,
        **kwargs,
    ):
        """
        Create a collection with:
          - id (VARCHAR primary key), text (VARCHAR)
          - one or more vector fields from `vector_config`
          - dynamic fields enabled (so `chunk.metadata` will be stored in $meta)
          - per-field indexes with metrics derived from `VectorConfig.distance`
        """

        client = self.get_client()
        if client.has_collection(collection_name):
            log.warning(
                f"Collection {collection_name} already exists, skipping creation"
            )
            return

        fields: list[FieldSchema] = [
            # Fixed 36-char UUIDv4 string (includes hyphens), used as primary key
            FieldSchema(
                name="id", dtype=DataType.VARCHAR, is_primary=True, max_length=36
            ),
            FieldSchema(name="text", dtype=DataType.VARCHAR, max_length=65535),
        ]

        # Build vector fields
        vector_names = set()
        for cfg in vector_config:
            if cfg.format == EmbeddingFormat.DENSE:
                fname = cfg.name
                if fname in vector_names:
                    raise ValueError(
                        "Each VectorConfig must have a unique 'name' for multi-vector collections"
                    )
                vector_names.add(fname)
                fields.append(
                    FieldSchema(
                        name=fname, dtype=DataType.FLOAT_VECTOR, dim=cfg.dimensions
                    )
                )
            elif cfg.format == EmbeddingFormat.SPARSE:
                fname = cfg.name
                if fname in vector_names:
                    raise ValueError(
                        "Each VectorConfig must have a unique 'name' for multi-vector collections"
                    )
                vector_names.add(fname)
                fields.append(
                    FieldSchema(
                        name=fname,
                        dtype=DataType.SPARSE_FLOAT_VECTOR,
                    )
                )
            else:
                raise ValueError(f"Unsupported embedding format: {cfg.format}")

        # Create the schema for the collection
        # TODO: add support for additional kwargs
        #  to pass to schema e.g. partition key, Milvus built in functions
        schema = CollectionSchema(
            fields,
            # default true for flexibility, Chunk's dynamic metadata will be stored into $meta
            # https://milvus.io/docs/enable-dynamic-field.md#Dynamic-Field
            enable_dynamic_field=True,
        )

        idx_params = client.prepare_index_params()
        if index_params is not None:
            idx_params = index_params
        # Add indexes
        if index_params is None:
            for cfg in vector_config:
                if cfg.format == EmbeddingFormat.DENSE:
                    fname = cfg.name
                    default_index_params = {
                        "field_name": fname,
                        "index_type": "AUTOINDEX",
                        "metric_type": self._metric_from_config(cfg),
                        "params": {},
                    }
                    idx_params.add_index(**default_index_params)
                elif cfg.format == EmbeddingFormat.SPARSE:
                    fname = cfg.name
                    default_index_params = {
                        "field_name": fname,
                        "index_type": "SPARSE_INVERTED_INDEX",
                        "metric_type": "IP",  # cosine not supported
                        "params": {},
                    }
                    idx_params.add_index(**default_index_params)

        # Create Collection
        client.create_collection(
            collection_name=collection_name,
            schema=schema,
            index_params=idx_params,
            **kwargs,
        )
        # load the collection by default
        client.load_collection(collection_name)

    def delete_collection(self, collection_name: str, **kwargs):
        client = self.get_client()
        if client.has_collection(collection_name):
            client.drop_collection(collection_name, **kwargs)

    def dump_collection(
        self,
        collection_name: str,
        page_size: int = 100,
    ) -> Generator[Chunk, None, None]:
        """
        Iterate all rows via offset paging.
        """
        client = self.get_client()
        offset = 0
        while True:
            res = client.query(
                collection_name=collection_name,
                filter="",
                output_fields=["*"],
                limit=page_size,
                offset=offset,
            )
            if not res:
                break
            for r in res:
                yield self._entity_to_chunk(r)
            if len(res) < page_size:
                break
            offset += page_size

    def prepare_index_params(self, **kwargs) -> IndexParams:
        client = self.get_client()
        return client.prepare_index_params(**kwargs)



import pytest

from datapizza.core.vectorstore import VectorConfig
from datapizza.type import EmbeddingFormat


def test_vector_config_model_post_init():
    with pytest.raises(ValueError):
        VectorConfig(name="dense_emb_name", format=EmbeddingFormat.DENSE)



import pytest

from datapizza.modules.parsers.docling.docling_parser import DoclingParser


@pytest.fixture
def mock_docling_parser(monkeypatch, tmp_path):
    """
    Fixture to create a DoclingParser instance with mocked converter behavior.
    """
    parser = DoclingParser(json_output_dir=str(tmp_path))

    # Mock converter and convert() result
    class MockResult:
        def __init__(self):
            self.document = self

        def export_to_dict(self):
            # Minimal fake Docling JSON
            return {
                "schema_name": "docling_test",
                "version": "1.0",
                "name": "mock_doc",
                "origin": "unit_test",
                "body": {"children": []},
            }

    class MockConverter:
        def convert(self, file_path):
            return MockResult()

    monkeypatch.setattr(parser, "_get_converter", lambda: MockConverter())
    return parser



import logging
import pickle

from datapizza.core.cache import Cache

import redis

log = logging.getLogger(__name__)


class RedisCache(Cache):
    """
    A Redis-based cache implementation.
    """

    def __init__(
        self, host="localhost", port=6379, db=0, expiration_time=3600
    ):  # 1 hour default
        self.redis = redis.Redis(host=host, port=port, db=db)
        self.expiration_time = expiration_time

    def get(self, key: str) -> str | None:
        """Retrieve and deserialize object"""
        pickled_obj = self.redis.get(key)
        if pickled_obj is None:
            return None
        return pickle.loads(pickled_obj)  # type: ignore

    def set(self, key: str, obj):
        """Serialize and store object"""
        pickled_obj = pickle.dumps(obj)
        self.redis.set(key, pickled_obj, ex=self.expiration_time)



from datapizza.core.modules.reranker import Reranker
from datapizza.type import Chunk


class CohereReranker(Reranker):
    """A reranker that uses the Cohere API to rerank documents."""

    def __init__(
        self,
        api_key: str,
        endpoint: str,
        top_n: int = 10,
        threshold: float | None = None,
        model: str = "model",
    ):
        """
        Args:
            api_key: The API key for the Cohere API.
            endpoint: The endpoint for the Cohere API.
            top_n: The number of documents to return.
            threshold: The threshold for the reranker.
        """

        self.api_key = api_key
        self.endpoint = endpoint
        self.top_n = top_n
        self.threshold = threshold
        self.model = model

        self.client = None
        self.a_client = None

    def _set_client(self):
        import cohere

        if not self.client:
            self.client = cohere.ClientV2(base_url=self.endpoint, api_key=self.api_key)

    def _set_a_client(self):
        import cohere

        if not self.a_client:
            self.a_client = cohere.AsyncClientV2(
                base_url=self.endpoint,
                api_key=self.api_key,
            )

    def _get_client(self):
        if not self.client:
            self._set_client()

        if not self.client:
            raise RuntimeError("Client not set")

        return self.client

    def _get_a_client(self):
        if not self.a_client:
            self._set_a_client()

        if not self.a_client:
            raise RuntimeError("Client not set")

        return self.a_client

    def rerank(self, query: str, documents: list[Chunk]) -> list[Chunk]:
        """
        Rerank documents based on query.

        Args:
            query: The query to rerank documents by.
            documents: The documents to rerank.

        Returns:
            The reranked documents.
        """
        client = self._get_client()

        response = client.rerank(
            model=self.model,
            query=query,
            documents=[single_document.text for single_document in documents],
            top_n=self.top_n,
        )

        reranked_documents: list[Chunk] = [None] * len(response.results)  # type: ignore
        filtered_documents = []

        for i, document in enumerate(response.results):
            index = document.index
            relevance_score = document.relevance_score

            # Apply threshold filtering if specified
            if self.threshold is None or relevance_score >= self.threshold:
                if self.threshold is None:
                    reranked_documents[i] = documents[index]
                else:
                    filtered_documents.append(documents[index])

        # Return filtered documents if threshold is applied, otherwise return reranked documents
        return filtered_documents if self.threshold is not None else reranked_documents

    async def a_rerank(self, query: str, documents: list[Chunk]) -> list[Chunk]:
        """
        Rerank documents based on query.

        Args:
            query: The query to rerank documents by.
            documents: The documents to rerank.

        Returns:
            The reranked documents.
        """
        client = self._get_a_client()

        response = await client.rerank(
            model=self.model,
            query=query,
            documents=[single_document.text for single_document in documents],
            top_n=self.top_n,
        )

        reranked_documents: list[Chunk] = [None] * len(response.results)  # type: ignore
        filtered_documents = []

        for i, document in enumerate(response.results):
            index = document.index
            relevance_score = document.relevance_score

            # Apply threshold filtering if specified
            if self.threshold is None or relevance_score >= self.threshold:
                if self.threshold is None:
                    reranked_documents[i] = documents[index]
                else:
                    filtered_documents.append(documents[index])

        # Return filtered documents if threshold is applied, otherwise return reranked documents
        return filtered_documents if self.threshold is not None else reranked_documents



import base64

from datapizza.memory.memory import Turn
from datapizza.memory.memory_adapter import MemoryAdapter
from datapizza.type import (
    ROLE,
    FunctionCallBlock,
    FunctionCallResultBlock,
    MediaBlock,
    StructuredBlock,
    TextBlock,
)

from google.genai import types


class GoogleMemoryAdapter(MemoryAdapter):
    def _turn_to_message(self, turn: Turn) -> dict:
        content = []
        for block in turn:
            block_dict = {}

            match block:
                case TextBlock():
                    block_dict = {"text": block.content}
                case FunctionCallBlock():
                    block_dict = {
                        "function_call": {"name": block.name, "args": block.arguments}
                    }
                case FunctionCallResultBlock():
                    block_dict = types.Part.from_function_response(
                        name=block.tool.name,
                        response={"result": block.result},
                    )
                case StructuredBlock():
                    block_dict = {"text": str(block.content)}
                case MediaBlock():
                    match block.media.media_type:
                        case "image":
                            block_dict = self._process_image_block(block)
                        case "pdf":
                            block_dict = self._process_pdf_block(block)

                        case "audio":
                            block_dict = self._process_audio_block(block)

                        case _:
                            raise NotImplementedError(
                                f"Unsupported media type: {block.media.media_type}"
                            )

            content.append(block_dict)

        return {
            "role": turn.role.google_role,
            "parts": (content),
        }

    def _process_audio_block(self, block: MediaBlock) -> types.Part:
        match block.media.source_type:
            case "raw":
                return types.Part.from_bytes(
                    data=block.media.source,
                    mime_type="audio/mp3",
                )

            case "path":
                with open(block.media.source, "rb") as f:
                    audio_bytes = f.read()

                return types.Part.from_bytes(
                    data=audio_bytes,
                    mime_type="audio/mp3",
                )

            case _:
                raise NotImplementedError(
                    f"Unsupported media source type: {block.media.source_type} for audio, source type supported: raw, path"
                )

    def _process_pdf_block(self, block: MediaBlock) -> types.Part | dict:
        match block.media.source_type:
            case "raw":
                return types.Part.from_bytes(
                    data=block.media.source,
                    mime_type="application/pdf",
                )
            case "base64":
                return {
                    "inline_data": {
                        "mime_type": "application/pdf",
                        "data": block.media.source,
                    }
                }
            case "path":
                with open(block.media.source, "rb") as f:
                    pdf_bytes = f.read()

                return {
                    "inline_data": {
                        "mime_type": "application/pdf",
                        "data": pdf_bytes,
                    }
                }

            case _:
                raise NotImplementedError(
                    f"Unsupported media source type: {block.media.source_type} only supported: raw, base64, path"
                )

    def _process_image_block(self, block: MediaBlock) -> dict:
        match block.media.source_type:
            case "url":
                return types.Part.from_uri(
                    file_uri=block.media.source,
                    mime_type=f"image/{block.media.extension}",
                )  # type: ignore
            case "base64":
                return {
                    "inline_data": {
                        "mime_type": f"image/{block.media.extension}",
                        "data": block.media.source,
                    }
                }
            case "path":
                with open(block.media.source, "rb") as image_file:
                    base64_image = base64.b64encode(image_file.read()).decode("utf-8")
                return {
                    "inline_data": {
                        "mime_type": f"image/{block.media.extension}",
                        "data": base64_image,
                    }
                }
            case _:
                raise NotImplementedError(
                    f"Unsupported media source type: {block.media.source_type} for image, only url, base64, path are supported"
                )

    def _text_to_message(self, text: str, role: ROLE) -> dict:
        return {"role": role.google_role, "parts": [{"text": text}]}



from collections.abc import AsyncIterator, Iterator
from typing import Any, Literal

from datapizza.core.cache import Cache
from datapizza.core.clients import Client, ClientResponse
from datapizza.core.clients.models import TokenUsage
from datapizza.memory import Memory
from datapizza.tools import Tool
from datapizza.type import FunctionCallBlock, TextBlock, ThoughtBlock

from anthropic import Anthropic, AsyncAnthropic

from .memory_adapter import AnthropicMemoryAdapter


class AnthropicClient(Client):
    """A client for interacting with the Anthropic API (Claude).

    This class provides methods for invoking the Anthropic API to generate responses
    based on given input data. It extends the Client class.
    """

    def __init__(
        self,
        api_key: str,
        model: str = "claude-3-5-sonnet-latest",
        system_prompt: str = "",
        temperature: float | None = None,
        cache: Cache | None = None,
    ):
        """
        Args:
            api_key: The API key for the Anthropic API.
            model: The model to use for the Anthropic API.
            system_prompt: The system prompt to use for the Anthropic API.
            temperature: The temperature to use for the Anthropic API.
            cache: The cache to use for the Anthropic API.
        """
        if temperature and not 0 <= temperature <= 2:
            raise ValueError("Temperature must be between 0 and 2")

        super().__init__(
            model_name=model,
            system_prompt=system_prompt,
            temperature=temperature,
            cache=cache,
        )
        self.api_key = api_key
        self.memory_adapter = AnthropicMemoryAdapter()
        self._set_client()

    def _set_client(self):
        if not self.client:
            self.client = Anthropic(api_key=self.api_key)

    def _set_a_client(self):
        if not self.a_client:
            self.a_client = AsyncAnthropic(api_key=self.api_key)

    def _convert_tools(self, tools: list[Tool]) -> list[dict[str, Any]]:
        """Convert tools to Anthropic tool format"""
        anthropic_tools = []
        for tool in tools:
            anthropic_tool = {
                "name": tool.name,
                "description": tool.description or "",
                "input_schema": {
                    "type": "object",
                    "properties": tool.properties,
                    "required": tool.required,
                },
            }
            anthropic_tools.append(anthropic_tool)
        return anthropic_tools

    def _convert_tool_choice(
        self, tool_choice: Literal["auto", "required", "none"] | list[str]
    ) -> dict | Literal["auto", "required", "none"]:
        if isinstance(tool_choice, list) and len(tool_choice) > 1:
            raise NotImplementedError(
                "multiple function names is not supported by Anthropic"
            )
        elif isinstance(tool_choice, list):
            return {
                "type": "tool",
                "name": tool_choice[0],
            }
        elif tool_choice == "required":
            return {"type": "any"}
        elif tool_choice == "auto":
            return {"type": "auto"}
        else:
            return tool_choice

    def _response_to_client_response(
        self, response, tool_map: dict[str, Tool] | None = None
    ) -> ClientResponse:
        """Convert Anthropic response to ClientResponse"""
        blocks = []

        if hasattr(response, "content") and response.content:
            if isinstance(
                response.content, list
            ):  # Claude 3 returns a list of content blocks
                for content_block in response.content:
                    if content_block.type == "text":
                        blocks.append(TextBlock(content=content_block.text))
                    elif content_block.type == "thinking":
                        # Summarized thinking content
                        blocks.append(ThoughtBlock(content=content_block.thinking))
                    elif content_block.type == "tool_use":
                        tool = tool_map.get(content_block.name) if tool_map else None
                        if not tool:
                            raise ValueError(f"Tool {content_block.name} not found")

                        blocks.append(
                            FunctionCallBlock(
                                id=content_block.id,
                                name=content_block.name,
                                arguments=content_block.input,
                                tool=tool,
                            )
                        )
            else:  # Handle as string for compatibility
                blocks.append(TextBlock(content=str(response.content)))

        stop_reason = response.stop_reason if hasattr(response, "stop_reason") else None

        return ClientResponse(
            content=blocks,
            stop_reason=stop_reason,
            usage=TokenUsage(
                prompt_tokens=response.usage.input_tokens,
                completion_tokens=response.usage.output_tokens,
                cached_tokens=response.usage.cache_read_input_tokens,
            ),
        )

    def _invoke(
        self,
        *,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int | None,
        system_prompt: str | None,
        **kwargs,
    ) -> ClientResponse:
        """Implementation of the abstract _invoke method for Anthropic"""
        if tools is None:
            tools = []
        client = self._get_client()
        messages = self._memory_to_contents(None, input, memory)
        # remove the model from the messages
        messages = [message for message in messages if message.get("role") != "model"]

        tool_map = {tool.name: tool for tool in tools}

        request_params = {
            "model": self.model_name,
            "messages": messages,
            "max_tokens": max_tokens or 2048,
            **kwargs,
        }

        if temperature:
            request_params["temperature"] = temperature

        if system_prompt:
            request_params["system"] = system_prompt

        if tools:
            request_params["tools"] = self._convert_tools(tools)
            request_params["tool_choice"] = self._convert_tool_choice(tool_choice)

        response = client.messages.create(**request_params)
        return self._response_to_client_response(response, tool_map)

    async def _a_invoke(
        self,
        *,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int | None,
        system_prompt: str | None,
        **kwargs,
    ) -> ClientResponse:
        if tools is None:
            tools = []
        client = self._get_a_client()
        messages = self._memory_to_contents(None, input, memory)
        # remove the model from the messages
        messages = [message for message in messages if message.get("role") != "model"]

        tool_map = {tool.name: tool for tool in tools}

        request_params = {
            "model": self.model_name,
            "messages": messages,
            "max_tokens": max_tokens or 2048,
            **kwargs,
        }

        if temperature:
            request_params["temperature"] = temperature

        if system_prompt:
            request_params["system"] = system_prompt

        if tools:
            request_params["tools"] = self._convert_tools(tools)
            request_params["tool_choice"] = self._convert_tool_choice(tool_choice)

        response = await client.messages.create(**request_params)
        return self._response_to_client_response(response, tool_map)

    def _stream_invoke(
        self,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int | None,
        system_prompt: str | None,
        **kwargs,
    ) -> Iterator[ClientResponse]:
        """Implementation of the abstract _stream_invoke method for Anthropic"""
        if tools is None:
            tools = []
        messages = self._memory_to_contents(None, input, memory)
        client = self._get_client()

        request_params = {
            "model": self.model_name,
            "messages": messages,
            "stream": True,
            "max_tokens": max_tokens or 2048,
            **kwargs,
        }

        if temperature:
            request_params["temperature"] = temperature

        if system_prompt:
            request_params["system"] = system_prompt

        if tools:
            request_params["tools"] = self._convert_tools(tools)
            request_params["tool_choice"] = self._convert_tool_choice(tool_choice)

        stream = client.messages.create(**request_params)

        input_tokens = 0
        output_tokens = 0
        message_text = ""
        thought_text = ""

        for chunk in stream:
            if (
                chunk.type == "content_block_delta"
                and hasattr(chunk, "delta")
                and chunk.delta
            ):
                if hasattr(chunk.delta, "text") and chunk.delta.text:
                    message_text += chunk.delta.text
                    yield ClientResponse(
                        content=[
                            ThoughtBlock(content=thought_text),
                            TextBlock(content=message_text),
                        ],
                        delta=chunk.delta.text,
                    )
                elif hasattr(chunk.delta, "thinking") and chunk.delta.thinking:
                    thought_text += chunk.delta.thinking

            if chunk.type == "message_start":
                input_tokens = (
                    chunk.message.usage.input_tokens if chunk.message.usage else 0
                )

            if chunk.type == "message_delta":
                output_tokens = max(
                    output_tokens, chunk.usage.output_tokens if chunk.usage else 0
                )

        yield ClientResponse(
            content=[
                ThoughtBlock(content=thought_text),
                TextBlock(content=message_text),
            ],
            delta="",
            stop_reason="end_turn",
            usage=TokenUsage(
                prompt_tokens=input_tokens,
                completion_tokens=output_tokens,
                cached_tokens=0,
            ),
        )

    async def _a_stream_invoke(
        self,
        input: str,
        tools: list[Tool] | None = None,
        memory: Memory | None = None,
        tool_choice: Literal["auto", "required", "none"] | list[str] = "auto",
        temperature: float | None = None,
        max_tokens: int | None = None,
        system_prompt: str | None = None,
        **kwargs,
    ) -> AsyncIterator[ClientResponse]:
        """Implementation of the abstract _a_stream_invoke method for Anthropic"""
        if tools is None:
            tools = []
        messages = self._memory_to_contents(None, input, memory)
        client = self._get_a_client()

        request_params = {
            "model": self.model_name,
            "messages": messages,
            "stream": True,
            "max_tokens": max_tokens or 2048,
            **kwargs,
        }
        if temperature:
            request_params["temperature"] = temperature
        if system_prompt:
            request_params["system"] = system_prompt

        if max_tokens:
            request_params["max_tokens"] = max_tokens

        if tools:
            request_params["tools"] = self._convert_tools(tools)
            request_params["tool_choice"] = self._convert_tool_choice(tool_choice)

        stream = await client.messages.create(**request_params)

        input_tokens = 0
        output_tokens = 0
        message_text = ""
        thought_text = ""

        async for chunk in stream:
            if (
                chunk.type == "content_block_delta"
                and hasattr(chunk, "delta")
                and chunk.delta
            ):
                if hasattr(chunk.delta, "text") and chunk.delta.text:
                    message_text += chunk.delta.text
                    yield ClientResponse(
                        content=[
                            ThoughtBlock(content=thought_text),
                            TextBlock(content=message_text),
                        ],
                        delta=chunk.delta.text,
                    )
                elif hasattr(chunk.delta, "thinking") and chunk.delta.thinking:
                    thought_text += chunk.delta.thinking

            if chunk.type == "message_start":
                input_tokens = (
                    chunk.message.usage.input_tokens if chunk.message.usage else 0
                )

            if chunk.type == "message_delta":
                output_tokens = max(
                    output_tokens, chunk.usage.output_tokens if chunk.usage else 0
                )

        yield ClientResponse(
            content=[
                ThoughtBlock(content=thought_text),
                TextBlock(content=message_text),
            ],
            delta="",
            stop_reason="end_turn",
            usage=TokenUsage(
                prompt_tokens=input_tokens,
                completion_tokens=output_tokens,
                cached_tokens=0,
            ),
        )

    def _structured_response(
        self,
        *args,
        **kwargs,
    ) -> ClientResponse:
        raise NotImplementedError("Anthropic does not support structured responses")

    async def _a_structured_response(self, *args, **kwargs):
        raise NotImplementedError("Anthropic does not support structured responses")



import contextlib
import uuid

import pytest
from datapizza.core.vectorstore import VectorConfig
from datapizza.type import (
    Chunk,
    EmbeddingFormat,
    SparseEmbedding,
)

from datapizza.vectorstores.milvus import MilvusVectorstore

MILVUS_URI = "./milvus.db"


def unique_coll(prefix="itest"):
    return f"{prefix}_{uuid.uuid4().hex[:8]}"


def make_chunk(cid: str, text: str = "hello", meta: bool = True) -> Chunk:
    metadata = {"source": "itest", "idx": cid} if meta else None
    return Chunk(
        id=cid,
        text=text,
        embeddings=[
            SparseEmbedding(
                name="sparse_vec",
                indices=[1, 5, 9],
                values=[0.5, 0.2, 0.1],
            )
        ],
        metadata=metadata,
    )


@pytest.fixture(scope="session")
def store():
    s = MilvusVectorstore(uri=MILVUS_URI)
    try:
        client = s.get_client()
        client.list_collections()
    except Exception as e:
        pytest.skip(f"Cannot connect to Milvus at {MILVUS_URI}: {e}")
    return s


@pytest.fixture
def vector_cfg():
    return [
        VectorConfig(
            name="sparse_vec",
            format=EmbeddingFormat.SPARSE,
            dimensions=1536,
            distance="Cosine",
        )
    ]


@pytest.fixture
def collection(store, vector_cfg):
    name = unique_coll()
    store.create_collection(
        collection_name=name,
        vector_config=vector_cfg,
    )
    yield name
    # Cleanup
    with contextlib.suppress(Exception):
        store.delete_collection(name)


def test_metric_from_config_sparse_cosine(store, vector_cfg):
    assert store._metric_from_config(vector_cfg[0]) == "COSINE"


def test_create_and_list_collections(store, vector_cfg):
    name = unique_coll()
    try:
        store.create_collection(name, vector_cfg)
        cols = store.get_collections()
        assert name in cols
    finally:
        store.delete_collection(name)


def test_add_and_retrieve_roundtrip(store, collection):
    c = make_chunk("id-1", text="alpha")
    store.add(c, collection_name=collection)
    # flush collection
    client = store.get_client()
    client.flush(collection_name=collection)
    out = store.retrieve(collection, ids=["id-1"])
    assert len(out) == 1
    got = out[0]
    assert got.id == "id-1"
    assert got.text == "alpha"
    assert got.metadata.get("source") == "itest"
    emb_names = {e.name for e in got.embeddings}
    assert "sparse_vec" in emb_names


def test_batch_add(store, collection):
    batch = [make_chunk(f"id-{i}") for i in range(6)]
    store.add(batch, collection_name=collection, batch_size=2)
    # flush collection
    client = store.get_client()
    client.flush(collection_name=collection)
    out = store.retrieve(collection, ids=[f"id-{i}" for i in range(6)])
    assert len(out) == 6


def test_update_upsert(store, collection):
    store.add(make_chunk("u1", text="v1"), collection_name=collection)
    store.update(collection, make_chunk("u1", text="v2"))
    got = store.retrieve(collection, ids=["u1"])[0]
    assert got.text == "v2"


def test_remove(store, collection):
    store.add(make_chunk("del-1"), collection_name=collection)
    store.remove(collection, ids=["del-1"])
    # flush collection
    client = store.get_client()
    client.flush(collection_name=collection)
    out = store.retrieve(collection, ids=["del-1"])
    assert out == []


def test_search_single_sparse(store, collection):
    store.add([make_chunk(f"s{i}") for i in range(5)], collection_name=collection)
    # flush collection
    client = store.get_client()
    client.flush(collection_name=collection)
    q = SparseEmbedding(name="sparse_vec", indices=[1, 2], values=[0.3, 0.7])
    res = store.search(collection, q, k=3)
    assert isinstance(res, list)
    assert len(res) == 3
    from datapizza.type import Chunk as _C

    assert all(isinstance(c, _C) for c in res)


def test_dump_collection_paged(store, collection):
    for i in range(7):
        store.add(make_chunk(f"d{i}"), collection_name=collection)
    # flush collection
    client = store.get_client()
    client.flush(collection_name=collection)
    dumped = list(store.dump_collection(collection, page_size=3))
    assert len(dumped) >= 7
    ids = {c.id for c in dumped}
    assert "d0" in ids and "d6" in ids



import json
from collections.abc import AsyncIterator, Iterator
from typing import Literal

import httpx
from datapizza.core.cache import Cache
from datapizza.core.clients import Client, ClientResponse
from datapizza.core.clients.models import TokenUsage
from datapizza.memory import Memory
from datapizza.tools import Tool
from datapizza.type import (
    FunctionCallBlock,
    Model,
    StructuredBlock,
    TextBlock,
    ThoughtBlock,
)

from openai import (
    AsyncOpenAI,
    AzureOpenAI,
    OpenAI,
)
from openai.types.responses import (
    ParsedResponseOutputMessage,
    ResponseCompletedEvent,
    ResponseFunctionToolCall,
    ResponseOutputMessage,
    ResponseReasoningItem,
    ResponseTextDeltaEvent,
)

from .memory_adapter import OpenAIMemoryAdapter


class OpenAIClient(Client):
    """A client for interacting with the OpenAI API.

    This class provides methods for invoking the OpenAI API to generate responses
    based on given input data. It extends the Client class.
    """

    def __init__(
        self,
        api_key: str,
        model: str = "gpt-4o-mini",
        system_prompt: str = "",
        temperature: float | None = None,
        cache: Cache | None = None,
        base_url: str | httpx.URL | None = None,
    ):
        """
        Args:
            api_key: The API key for the OpenAI API.
            model: The model to use for the OpenAI API.
            system_prompt: The system prompt to use for the OpenAI API.
            temperature: The temperature to use for the OpenAI API.
            cache: The cache to use for the OpenAI API.
            base_url: The base URL for the OpenAI API.
        """

        if temperature and not 0 <= temperature <= 2:
            raise ValueError("Temperature must be between 0 and 2")

        super().__init__(
            model_name=model,
            system_prompt=system_prompt,
            temperature=temperature,
            cache=cache,
        )

        self.base_url = base_url
        self.api_key = api_key
        self.memory_adapter = OpenAIMemoryAdapter()
        self._set_client()

    def _set_client(self):
        if not self.client:
            self.client = OpenAI(api_key=self.api_key, base_url=self.base_url)

    def _set_a_client(self):
        if not self.a_client:
            self.a_client = AsyncOpenAI(api_key=self.api_key, base_url=self.base_url)

    def _response_to_client_response(
        self, response, tool_map: dict[str, Tool] | None
    ) -> ClientResponse:
        blocks = []

        # Handle new response format with direct content array
        if hasattr(response, "output_parsed"):
            blocks.append(StructuredBlock(content=response.output_parsed))

        if hasattr(response, "output") and response.output:
            for content_item in response.output:
                if isinstance(content_item, ResponseOutputMessage) and not isinstance(
                    content_item, ParsedResponseOutputMessage
                ):
                    for content in content_item.content:
                        if content.type == "output_text":
                            blocks.append(TextBlock(content=content.text))
                elif isinstance(content_item, ResponseReasoningItem):
                    if content_item.summary:
                        blocks.append(
                            ThoughtBlock(content=content_item.summary[0].text)
                        )

                elif isinstance(content_item, ResponseFunctionToolCall):
                    if not tool_map:
                        raise ValueError("Tool map is required")

                    tool = tool_map.get(content_item.name)
                    if not tool:
                        raise ValueError(f"Tool {content_item.name} not found")
                    blocks.append(
                        FunctionCallBlock(
                            id=content_item.call_id,
                            name=content_item.name,
                            arguments=json.loads(content_item.arguments)
                            if isinstance(content_item.arguments, str)
                            else content_item.arguments,
                            tool=tool,
                        )
                    )

        # Handle usage from new format
        usage = getattr(response, "usage", None)
        if usage:
            prompt_tokens = getattr(usage, "input_tokens", 0)
            completion_tokens = getattr(usage, "output_tokens", 0)
            cached_tokens = 0
            # Handle input_tokens_details for cached tokens
            if hasattr(usage, "input_tokens_details") and usage.input_tokens_details:
                cached_tokens = getattr(usage.input_tokens_details, "cached_tokens", 0)

        # Handle stop reason - use status from new format
        stop_reason = getattr(response, "status", None)
        if not stop_reason and hasattr(response, "choices") and response.choices:
            stop_reason = response.choices[0].finish_reason

        return ClientResponse(
            content=blocks,
            stop_reason=stop_reason,
            usage=TokenUsage(
                prompt_tokens=prompt_tokens or 0,
                completion_tokens=completion_tokens or 0,
                cached_tokens=cached_tokens or 0,
            ),
        )

    def _convert_tools(self, tool: Tool) -> dict:
        """Convert tools to OpenAI function format"""
        return {
            "type": "function",
            "name": tool.name,
            "description": tool.description,
            "parameters": {
                "type": "object",
                "properties": tool.properties,
                "required": tool.required,
            },
        }

    def _convert_tool_choice(
        self, tool_choice: Literal["auto", "required", "none"] | list[str]
    ) -> dict | Literal["auto", "required", "none"]:
        if isinstance(tool_choice, list) and len(tool_choice) > 1:
            raise NotImplementedError(
                "multiple function names is not supported by OpenAI"
            )
        elif isinstance(tool_choice, list):
            return {
                "type": "function",
                "name": tool_choice[0],
            }
        else:
            return tool_choice

    def _invoke(
        self,
        *,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        **kwargs,
    ) -> ClientResponse:
        if tools is None:
            tools = []
        messages = self._memory_to_contents(system_prompt, input, memory)

        tool_map = {tool.name: tool for tool in tools}

        kwargs = {
            "model": self.model_name,
            "input": messages,
            "stream": False,
            "max_output_tokens": max_tokens,
            **kwargs,
        }
        if temperature:
            kwargs["temperature"] = temperature

        if tools:
            kwargs["tools"] = [self._convert_tools(tool) for tool in tools]
            kwargs["tool_choice"] = self._convert_tool_choice(tool_choice)

        client: OpenAI = self._get_client()
        response = client.responses.create(**kwargs)
        return self._response_to_client_response(response, tool_map)

    async def _a_invoke(
        self,
        *,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        **kwargs,
    ) -> ClientResponse:
        if tools is None:
            tools = []
        messages = self._memory_to_contents(system_prompt, input, memory)

        tool_map = {tool.name: tool for tool in tools}

        kwargs = {
            "model": self.model_name,
            "input": messages,
            "stream": False,
            "max_output_tokens": max_tokens,
            **kwargs,
        }
        if temperature:
            kwargs["temperature"] = temperature

        if tools:
            kwargs["tools"] = [self._convert_tools(tool) for tool in tools]
            kwargs["tool_choice"] = self._convert_tool_choice(tool_choice)

        a_client = self._get_a_client()
        response = await a_client.responses.create(**kwargs)
        return self._response_to_client_response(response, tool_map)

    def _stream_invoke(
        self,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        **kwargs,
    ) -> Iterator[ClientResponse]:
        if tools is None:
            tools = []
        messages = self._memory_to_contents(system_prompt, input, memory)

        tool_map = {tool.name: tool for tool in tools}

        kwargs = {
            "model": self.model_name,
            "input": messages,
            "stream": True,
            "max_output_tokens": max_tokens,
            # "stream_options": {"include_usage": True},
            **kwargs,
        }
        if temperature:
            kwargs["temperature"] = temperature

        if tools:
            kwargs["tools"] = [self._convert_tools(tool) for tool in tools]
            kwargs["tool_choice"] = self._convert_tool_choice(tool_choice)

        response = self.client.responses.create(**kwargs)
        for chunk in response:
            if isinstance(chunk, ResponseTextDeltaEvent):
                yield ClientResponse(
                    content=[],
                    delta=chunk.delta,
                    stop_reason=None,
                    usage=TokenUsage(
                        prompt_tokens=0,
                        completion_tokens=0,
                        cached_tokens=0,
                    ),
                )

            if isinstance(chunk, ResponseCompletedEvent):
                yield self._response_to_client_response(chunk.response, tool_map)

    async def _a_stream_invoke(
        self,
        input: str,
        tools: list[Tool] | None = None,
        memory: Memory | None = None,
        tool_choice: Literal["auto", "required", "none"] | list[str] = "auto",
        temperature: float | None = None,
        max_tokens: int | None = None,
        system_prompt: str | None = None,
        **kwargs,
    ) -> AsyncIterator[ClientResponse]:
        if tools is None:
            tools = []
        messages = self._memory_to_contents(system_prompt, input, memory)

        tool_map = {tool.name: tool for tool in tools}
        kwargs = {
            "model": self.model_name,
            "input": messages,
            "stream": True,
            "max_output_tokens": max_tokens,
            # "stream_options": {"include_usage": True},
            **kwargs,
        }
        if temperature:
            kwargs["temperature"] = temperature

        if tools:
            kwargs["tools"] = [self._convert_tools(tool) for tool in tools]
            kwargs["tool_choice"] = self._convert_tool_choice(tool_choice)

        a_client = self._get_a_client()

        async for chunk in await a_client.responses.create(**kwargs):
            if isinstance(chunk, ResponseTextDeltaEvent):
                yield ClientResponse(
                    content=[],
                    delta=chunk.delta,
                    stop_reason=None,
                    usage=TokenUsage(
                        prompt_tokens=0,
                        completion_tokens=0,
                        cached_tokens=0,
                    ),
                )

            if isinstance(chunk, ResponseCompletedEvent):
                yield self._response_to_client_response(chunk.response, tool_map)

    def _structured_response(
        self,
        input: str,
        output_cls: type[Model],
        memory: Memory | None,
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        tools: list[Tool] | None,
        tool_choice: Literal["auto", "required", "none"] | list[str] = "auto",
        **kwargs,
    ) -> ClientResponse:
        # Add system message to enforce JSON output

        if tools is None:
            tools = []

        messages = self._memory_to_contents(system_prompt, input, memory)

        tool_map = {tool.name: tool for tool in tools}
        kwargs = {
            "model": self.model_name,
            "input": messages,
            "text_format": output_cls,
            "max_output_tokens": max_tokens,
            **kwargs,
        }
        if temperature:
            kwargs["temperature"] = temperature

        if tools:
            kwargs["tools"] = [self._convert_tools(tool) for tool in tools]
            kwargs["tool_choice"] = self._convert_tool_choice(tool_choice)
            # Structured response needs strict mode and no additional properties
            for tool in kwargs["tools"]:
                tool["function"]["strict"] = True
                tool["function"]["parameters"]["additionalProperties"] = False

        response = self.client.responses.parse(**kwargs)

        return self._response_to_client_response(response, tool_map)

    async def _a_structured_response(
        self,
        input: str,
        output_cls: type[Model],
        memory: Memory | None,
        temperature: float,
        max_tokens: int,
        system_prompt: str | None = None,
        tools: list[Tool] | None = None,
        tool_choice: Literal["auto", "required", "none"] | list[str] = "auto",
        **kwargs,
    ):
        if tools is None:
            tools = []

        messages = self._memory_to_contents(system_prompt, input, memory)
        tool_map = {tool.name: tool for tool in tools}

        kwargs = {
            "model": self.model_name,
            "input": messages,
            "text_format": output_cls,
            "max_output_tokens": max_tokens,
            **kwargs,
        }
        if temperature:
            kwargs["temperature"] = temperature

        if tools:
            kwargs["tools"] = [self._convert_tools(tool) for tool in tools]
            kwargs["tool_choice"] = self._convert_tool_choice(tool_choice)
            # Structured response needs strict mode and no additional properties
            for tool in kwargs["tools"]:
                tool["function"]["strict"] = True
                tool["function"]["parameters"]["additionalProperties"] = False

        a_client = self._get_a_client()
        response = await a_client.responses.parse(**kwargs)

        return self._response_to_client_response(response, tool_map)

    def _embed(
        self, text: str | list[str], model_name: str | None, **kwargs
    ) -> list[float] | list[list[float]]:
        """Embed a text using the model"""
        response = self.client.embeddings.create(
            input=text, model=model_name or self.model_name, **kwargs
        )

        embeddings = [item.embedding for item in response.data]

        if isinstance(text, str):
            return embeddings[0] if embeddings else []

        return embeddings or []

    async def _a_embed(
        self, text: str | list[str], model_name: str | None, **kwargs
    ) -> list[float] | list[list[float]]:
        """Embed a text using the model"""

        a_client = self._get_a_client()
        response = await a_client.embeddings.create(
            input=text, model=model_name or self.model_name, **kwargs
        )

        embeddings = [item.embedding for item in response.data]

        if isinstance(text, str):
            return embeddings[0] if embeddings else []

        return embeddings or []

    def _is_azure_client(self) -> bool:
        return isinstance(self._get_client(), AzureOpenAI)



import pytest
from pydantic import BaseModel

from datapizza.clients.mock_client import MockClient
from datapizza.core.clients import Client, ClientResponse
from datapizza.memory.memory import Memory
from datapizza.type import TextBlock
from datapizza.type.type import ROLE


def test_cannot_instantiate_abstract_class():
    with pytest.raises(TypeError):
        Client("a", "b", 0.5)


def test_mock_client():
    llm = MockClient(model_name="mock_client")

    response = ClientResponse(content=[TextBlock(content="Mock response")])

    assert llm.invoke(input="Mock response") == response


def test_stream_response():
    llm = MockClient(model_name="mock_client")
    text = "Here a long text to return"
    response = llm.stream_invoke(input=text)

    final_response = ""
    for i, res in enumerate(response):
        final_response += res.delta
        assert res.delta == text[i]
        assert res.text == text[: i + 1]

    assert final_response == text
    assert res.text == text


def test_memory():
    memory = Memory()
    textBlock = TextBlock(content="Mock response")
    textBlock2 = TextBlock(content="Mock response 2")
    memory.add_turn(textBlock, role=ROLE.ASSISTANT)
    assert memory[0].blocks == [textBlock]

    memory.clear()
    assert len(memory) == 0

    memory.add_turn([textBlock], role=ROLE.ASSISTANT)
    assert memory[0].blocks == [textBlock]

    memory.clear()
    assert len(memory) == 0

    memory.new_turn()
    assert len(memory) == 1

    memory.add_to_last_turn(textBlock)
    assert memory[0].blocks == [textBlock]

    memory.add_to_last_turn(textBlock2)
    assert memory[0].blocks == [textBlock, textBlock2]

    assert len(memory) == 1


def test_client_with_memory():
    text1 = "Mock response"
    text2 = "secondresponse"

    # Creazione Memory
    memory = Memory()

    # Creazione client
    llm = MockClient(model_name="mock_client")
    response = llm.invoke(input=text1, memory=memory)

    assert response.content == [TextBlock(content=text1)]

    assert response.text == text1

    memory.add_turn(response.content, role=ROLE.ASSISTANT)

    response = llm.invoke(input=text2, memory=memory)
    assert response.text == text1 + text2


def test_response_format():
    class TestModel(BaseModel):
        test: int

    input = '{"test": 1}'

    llm = MockClient(model_name="mock_client")
    response = llm.structured_response(input=input, output_cls=TestModel)

    # assert isinstance(response, TestModel)
    # assert response == TestModel(test=1)

    assert response.structured_data == [TestModel(test=1)]


def test_response_first_text():
    response = ClientResponse(
        content=[TextBlock(content="string1"), TextBlock("text2")]
    )
    assert response.first_text == "string1"


def test_client_module():
    llm = MockClient(model_name="mock_client")
    llm_module = llm.as_module_component()
    response = llm_module._run(input="Mock response")
    assert isinstance(response, ClientResponse)
    assert response.content == [TextBlock(content="Mock response")]

    memory = Memory()
    memory.add_turn(TextBlock(content="old_message"), role=ROLE.ASSISTANT)
    response_with_memory = llm_module._run(input="Mock response", memory=memory)
    assert "old_message" in response_with_memory.text
    assert "Mock response" in response_with_memory.text



import logging
from collections.abc import Generator
from typing import Any

from datapizza.core.vectorstore import VectorConfig, Vectorstore
from datapizza.type import (
    Chunk,
    DenseEmbedding,
    Embedding,
    EmbeddingFormat,
    SparseEmbedding,
)
from qdrant_client import AsyncQdrantClient, QdrantClient, models

log = logging.getLogger(__name__)


class QdrantVectorstore(Vectorstore):
    """
    datapizza-ai implementation of a Qdrant vectorstore.
    """

    def __init__(
        self,
        host: str | None = None,
        port: int = 6333,
        api_key: str | None = None,
        **kwargs,
    ):
        """
        Initialize the QdrantVectorstore.

        Args:
            host (str, optional): The host to use for the Qdrant client. Defaults to None.
            port (int, optional): The port to use for the Qdrant client. Defaults to 6333.
            api_key (str, optional): The API key to use for the Qdrant client. Defaults to None.
            **kwargs: Additional keyword arguments to pass to the Qdrant client.
        """
        if host is None and "location" not in kwargs:
            raise ValueError("Either host or location must be provided")

        self.client: QdrantClient
        self.a_client: AsyncQdrantClient
        self.batch_size: int = 100
        self.host: str | None = host
        self.port: int = port
        self.api_key: str | None = api_key
        self.kwargs: dict[str, Any] = kwargs

    def get_client(self) -> QdrantClient:
        if not hasattr(self, "client"):
            self._init_client()
        return self.client

    def _get_a_client(self) -> AsyncQdrantClient:
        if not hasattr(self, "a_client"):
            self._init_a_client()
        return self.a_client

    def _init_client(self):
        self.client = QdrantClient(
            host=self.host, port=self.port, api_key=self.api_key, **self.kwargs
        )

    def _init_a_client(self):
        self.a_client = AsyncQdrantClient(
            host=self.host, port=self.port, api_key=self.api_key, **self.kwargs
        )

    def add(self, chunk: Chunk | list[Chunk], collection_name: str | None = None):
        """Add a single chunk or list of chunks to the vectorstore.
        Args:
            chunk (Chunk | list[Chunk]): The chunk or list of chunks to add.
            collection_name (str, optional): The name of the collection to add the chunks to. Defaults to None.
        """
        client = self.get_client()

        if not collection_name:
            raise ValueError("Collection name must be set")

        chunks = [chunk] if isinstance(chunk, Chunk) else chunk
        points = []

        for chunk in chunks:
            points.append(self._process_chunk(chunk))

        # TODO: Process in batches
        for p in points:
            try:
                client.upsert(collection_name=collection_name, points=[p], wait=True)
            except Exception as e:
                log.error(f"Failed to add points to Qdrant: {e!s}")
                raise e

    async def a_add(
        self, chunk: Chunk | list[Chunk], collection_name: str | None = None
    ):
        client = self._get_a_client()

        if not collection_name:
            raise ValueError("Collection name must be set")

        chunks = [chunk] if isinstance(chunk, Chunk) else chunk
        points = []

        for chunk in chunks:
            points.append(self._process_chunk(chunk))

        # TODO: Process in batches
        for p in points:
            try:
                await client.upsert(
                    collection_name=collection_name, points=[p], wait=True
                )
            except Exception as e:
                log.error(f"Failed to add points to Qdrant: {e!s}")
                raise e

    def _process_chunk(self, chunk: Chunk) -> models.PointStruct:
        """Process a chunk into a Qdrant point."""
        if not chunk.embeddings:
            raise ValueError("Chunk must have an embedding")

        vector = {}
        for v in chunk.embeddings:
            if isinstance(v, DenseEmbedding):
                vector[v.name] = v.vector
            elif isinstance(v, SparseEmbedding):
                vector[v.name] = models.SparseVector(values=v.values, indices=v.indices)
            else:
                raise ValueError(f"Unsupported embedding type: {type(v)}")

        return models.PointStruct(
            id=str(chunk.id),
            payload={
                "text": chunk.text,
                **chunk.metadata,
            },
            vector=vector,  # type: ignore
        )

    def update(self, collection_name: str, payload: dict, points: list[int], **kwargs):
        client = self.get_client()
        client.overwrite_payload(
            collection_name=collection_name,
            payload=payload,
            points=points,  # type: ignore
            **kwargs,
        )

    def retrieve(self, collection_name: str, ids: list[str], **kwargs) -> list[Chunk]:
        """Retrieve chunks from a collection by their IDs.
        Args:
            collection_name (str): The name of the collection to retrieve the chunks from.
            ids (list[str]): The IDs of the chunks to retrieve.
            **kwargs: Additional keyword arguments to pass to the Qdrant client.
        Returns:
            list[Chunk]: The list of chunks retrieved from the collection.
        """
        client = self.get_client()
        return self._point_to_chunk(
            client.retrieve(
                collection_name=collection_name,
                ids=ids,
                **kwargs,
            )
        )

    def remove(self, collection_name: str, ids: list[str], **kwargs):
        """Remove chunks from a collection by their IDs.
        Args:
            collection_name (str): The name of the collection to remove the chunks from.
            ids (list[str]): The IDs of the chunks to remove.
            **kwargs: Additional keyword arguments to pass to the Qdrant client.
        """
        client = self.get_client()
        client.delete(
            collection_name=collection_name,
            points_selector=models.PointIdsList(
                points=ids,  # type: ignore
            ),
            **kwargs,
        )

    def search(
        self,
        collection_name: str,
        query_vector: list[float] | SparseEmbedding | dict,
        k: int = 10,
        vector_name: str | None = None,
        **kwargs,
    ) -> list[Chunk]:
        """
        Search for chunks in a collection by their query vector.

        Args:
            collection_name (str): The name of the collection to search in.
            query_vector (list[float]): The query vector to search for.
            k (int, optional): The number of results to return. Defaults to 10.
            vector_name (str, optional): The name of the vector to search for. Defaults to None.
            **kwargs: Additional keyword arguments to pass to the Qdrant client.

        Returns:
            list[Chunk]: The list of chunks found in the collection.
        """
        client = self.get_client()

        if isinstance(query_vector, list) and all(
            isinstance(v, float) for v in query_vector
        ):
            if not vector_name:
                vectors_config = list(
                    client.get_collection(collection_name).config.params.vectors
                )
                if vectors_config and len(vectors_config) > 1:
                    raise ValueError(
                        f"Vector name not specified and multiple dense vectors are configured. Available vector names: {vectors_config}"
                    )
                vector_name = str(vectors_config[0])
            qry = (vector_name, query_vector) if vector_name else query_vector

        elif isinstance(query_vector, dict):
            indices = query_vector.get("indices", [])
            values = query_vector.get("values", [])
            qry = models.NamedSparseVector(
                name=vector_name or "default",
                vector=models.SparseVector(indices=indices, values=values),
            )

        elif isinstance(query_vector, SparseEmbedding):
            qry = models.NamedSparseVector(
                name=query_vector.name,
                vector=models.SparseVector(
                    indices=query_vector.indices, values=query_vector.values
                ),
            )
        else:
            raise ValueError(f"Unsupported query vector type: {type(query_vector)}")

        hits = client.search(
            collection_name=collection_name,
            query_vector=qry,
            limit=k,  # Return k closest points
            **kwargs,
        )
        return self._point_to_chunk(hits)

    async def a_search(
        self,
        collection_name: str,
        query_vector: list[float],
        k: int = 10,
        vector_name: str | None = None,
        **kwargs,
    ) -> list[Chunk]:
        """Search for chunks in a collection by their query vector."""
        client = self._get_a_client()

        qry = (vector_name, query_vector) if vector_name else query_vector

        hits = await client.search(
            collection_name=collection_name,
            query_vector=qry,
            limit=k,  # Return k closest points
            **kwargs,
        )
        return self._point_to_chunk(hits)

    def get_collections(self):
        """Get all collections in Qdrant."""
        client = self.get_client()
        return client.get_collections()

    def create_collection(
        self, collection_name: str, vector_config: list[VectorConfig], **kwargs
    ):
        """Create a new collection in Qdrant if it doesn't exist with the specified vector configurations

        Args:
            collection_name: Name of the collection to create
            vector_config: List of vector configurations specifying dimensions and distance metrics
            **kwargs: Additional arguments to pass to Qdrant's create_collection
        """

        client = self.get_client()

        if client.collection_exists(collection_name):
            log.warning(
                f"Collection {collection_name} already exists, skipping creation"
            )
            return

        sparse_config: (
            dict[str, models.SparseVectorParams] | models.SparseVectorParams | None
        ) = None
        config = None
        try:
            config = {
                v.name: models.VectorParams(
                    size=v.dimensions,  # type: ignore
                    distance=v.distance.value,  # type: ignore
                )
                for v in vector_config
                if v.format == EmbeddingFormat.DENSE
            }
            sparse_config = {
                v.name: models.SparseVectorParams()
                for v in vector_config
                if v.format == EmbeddingFormat.SPARSE
            }

            client.create_collection(
                collection_name=collection_name,
                vectors_config=config,
                sparse_vectors_config=sparse_config,
                **kwargs,
            )
        except Exception as e:
            log.error(f"Failed to create collection {collection_name}: {e!s}")
            raise e

    def delete_collection(self, collection_name: str, **kwargs):
        """Delete a collection in Qdrant."""
        client = self.get_client()
        client.delete_collection(collection_name=collection_name, **kwargs)

    def dump_collection(
        self,
        collection_name: str,
        page_size: int = 100,
        with_vectors: bool = False,
    ) -> Generator[Chunk, None, None]:
        """
        Dumps all points from a collection in a chunk-wise manner.

        Args:
            collection_name: Name of the collection to dump.
            page_size: Number of points to retrieve per batch.
            with_vectors: Whether to include vectors in the dumped chunks.

        Yields:
            Chunk: A chunk object from the collection.
        """
        client = self.get_client()
        next_page_offset = None

        while True:
            points, next_page_offset = client.scroll(
                collection_name=collection_name,
                limit=page_size,
                offset=next_page_offset,
                with_payload=True,
                with_vectors=with_vectors,
            )

            if not points:
                break

            yield from self._point_to_chunk(points)

            if next_page_offset is None:
                break

    def _point_to_chunk(self, points) -> list[Chunk]:
        """
        Convert Qdrant points to Chunk objects.

        Args:
            points: List of Qdrant point objects

        Returns:
            List of Chunk objects with appropriate embeddings
        """
        chunks = []

        for point in points:
            vector = point.vector
            embeddings: list[Embedding] = []

            # Handle dictionary of named vectors
            if isinstance(vector, dict):
                for name, vec in vector.items():
                    if isinstance(vec, models.SparseVector):
                        embeddings.append(
                            SparseEmbedding(
                                name=name,
                                values=vec.values,
                                indices=vec.indices,
                            )
                        )
                    elif isinstance(vec, list):
                        embeddings.append(DenseEmbedding(name=name, vector=vec))
            # Handle single dense vector (list)
            elif isinstance(vector, list):
                embeddings.append(DenseEmbedding(name="dense", vector=vector))
            # Handle single sparse vector
            elif isinstance(vector, models.SparseVector):
                embeddings.append(
                    SparseEmbedding(
                        name="sparse", values=vector.values, indices=vector.indices
                    )
                )
            elif vector is None:
                embeddings = []
            else:
                raise ValueError(f"Unsupported vector type: {type(vector)}")

            chunks.append(
                Chunk(
                    id=point.id,
                    metadata=point.payload,
                    text=point.payload["text"],
                    embeddings=embeddings,
                )
            )

        return chunks



def extract_media_from_docling_bbox(
    bbox: dict,
    file_path: str,
    page_number: int,
    *,
    zoom: float = 2.0,
) -> str:
    """Extract a base64 PNG crop from a PDF using a Docling-style bbox.

    The bbox is expected to be a dict with keys:
    - l, t, r, b: floats in PDF coordinate space
    - coord_origin: "BOTTOMLEFT" (Docling default) or "TOPLEFT" (string or enum)
    """
    try:
        import fitz
    except ImportError as e:
        raise ImportError(
            "PyMuPDF is not installed. Please install it using `pip install PyMuPDF`."
        ) from e

    try:
        from PIL import Image
    except ImportError as e:
        raise ImportError(
            "PIL is not installed. Please install it using `pip install Pillow`."
        ) from e

    if not isinstance(bbox, dict):
        raise ValueError("bbox must be a dict with keys l,t,r,b,coord_origin")

    l = float(bbox.get("l", 0.0))
    r = float(bbox.get("r", 0.0))
    t = float(bbox.get("t", 0.0))
    b = float(bbox.get("b", 0.0))

    # Normalize coord origin: accept enum (with .name), strings like
    # "CoordOrigin.TOPLEFT", "TOP_LEFT", etc.
    origin_val = bbox.get("coord_origin", "BOTTOMLEFT")
    if hasattr(origin_val, "name"):
        origin_str = str(getattr(origin_val, "name", "")).upper()
    else:
        origin_str = str(origin_val).upper()
    if "." in origin_str:
        origin_str = origin_str.split(".")[-1]
    origin_str = origin_str.replace("_", "").replace("-", "")
    if origin_str not in {"BOTTOMLEFT", "TOPLEFT"}:
        raise ValueError(f"Unsupported coord_origin: {origin_val!r}")

    with fitz.open(file_path) as doc:
        page = doc[page_number - 1]
        mat = fitz.Matrix(zoom, zoom)
        pix = page.get_pixmap(matrix=mat)  # type: ignore

        img = Image.frombytes("RGB", (pix.width, pix.height), pix.samples)  # type: ignore
        x_scale = pix.width / page.rect.width
        y_scale = pix.height / page.rect.height

        if origin_str == "BOTTOMLEFT":
            left = l * x_scale
            right = r * x_scale
            top = pix.height - (t * y_scale)
            bottom = pix.height - (b * y_scale)
        else:
            left = l * x_scale
            right = r * x_scale
            top = t * y_scale
            bottom = b * y_scale

        left_i = max(0, min(round(left), img.width))
        right_i = max(0, min(round(right), img.width))
        top_i = max(0, min(round(top), img.height))
        bottom_i = max(0, min(round(bottom), img.height))

        # Ensure proper ordering after rounding/clamping
        left_i, right_i = (left_i, right_i) if left_i <= right_i else (right_i, left_i)
        top_i, bottom_i = (top_i, bottom_i) if top_i <= bottom_i else (bottom_i, top_i)

        if right_i - left_i <= 1 or bottom_i - top_i <= 1:
            raise ValueError("Invalid crop region computed from bbox")

        crop = img.crop((left_i, top_i, right_i, bottom_i))

        import base64
        import io

        buf = io.BytesIO()
        crop.save(buf, format="PNG")
        return base64.b64encode(buf.getvalue()).decode("utf-8")



from jinja2 import Template

from datapizza.core.modules.prompt import Prompt
from datapizza.core.utils import extract_media
from datapizza.memory import Memory
from datapizza.type import ROLE, Block, Chunk, Media, MediaBlock, TextBlock


class ImageRAGPrompt(Prompt):
    """
    Create a memory for a image RAG system.

    """

    def __init__(
        self,
        user_prompt_template: str,
        image_prompt_presentation: str,
        each_image_template: str,
    ):
        """
        Args:
            user_prompt_template: str # The user prompt jinja template
            image_prompt_presentation: str # The image prompt jinja template
            each_image_template: str # The each image jinja template
        """
        self.user_prompt_template = Template(user_prompt_template)
        self.image_prompt_presentation = image_prompt_presentation
        self.each_image_template = Template(each_image_template)

    def _extract_images_from_chunk(self, chunk: Chunk) -> list[MediaBlock]:
        metadata = chunk.metadata
        list_bboxes = metadata["boundingRegions"]
        path_pdf = metadata["document_name"]

        list_media_blocks: list[MediaBlock] = []

        for bbox in list_bboxes:
            page_number = bbox["pageNumber"]
            polygon = bbox["polygon"]

            # Extract the image from the PDF
            image_as_base64 = extract_media(
                coordinates=polygon,
                file_path=path_pdf,
                page_number=page_number,
            )

            media = Media(
                media_type="image",
                source=image_as_base64,
                source_type="base64",
                extension="png",
            )

            list_media_blocks.append(MediaBlock(media=media))

        return list_media_blocks

    def format(
        self,
        chunks: list[Chunk],
        user_query: str,
        retrieval_query: str,
        memory: Memory | None = None,
    ) -> Memory:
        """
        Creates a new memory object that includes:
        - Existing memory messages
        - User's query
        - Function call retrieval results
        - Chunks retrieval results

        Args:
            chunks: The chunks to add to the memory.
            user_query: The user's query.
            retrieval_query: The query to search the vectorstore for.
            memory: The memory object to add the new messages to.

        Returns:
            memory: A new memory object with the new messages.
        """
        new_memory = Memory()

        if memory:
            for turn in memory:
                new_memory.add_turn(turn.blocks, turn.role)

        all_blocks: list[Block] = [TextBlock(content=self.image_prompt_presentation)]

        for chunk in chunks:
            all_blocks.extend(
                [
                    TextBlock(
                        content=self.each_image_template.render(
                            path_pdf=chunk.metadata["document_name"].split("/")[-1],
                        )
                    )
                ]
            )
            all_blocks.extend(self._extract_images_from_chunk(chunk))

        formatted_user_prompt = self.user_prompt_template.render(
            user_prompt=user_query, retrieval_query=retrieval_query
        )

        all_blocks.append(TextBlock(content=formatted_user_prompt))

        new_memory.add_turn(blocks=all_blocks, role=ROLE.USER)

        return new_memory



import hashlib
import json

from datapizza.type import ROLE, Block, FunctionCallBlock


class Turn:
    def __init__(
        self,
        blocks: list[Block],
        role: ROLE = ROLE.ASSISTANT,
    ):
        if not isinstance(blocks, list):
            raise ValueError("blocks must be a list")
        if not all(isinstance(block, Block) for block in blocks):
            raise ValueError("all items in blocks must be Block instances")

        self.blocks = blocks
        self.role = role

    def __iter__(self):
        return iter(self.blocks)

    def __len__(self):
        return len(self.blocks)

    def __getitem__(self, index):
        return self.blocks[index]

    def __setitem__(self, index, value):
        self.blocks[index] = value

    def __delitem__(self, index):
        del self.blocks[index]

    def append(self, block: Block):
        self.blocks.append(block)

    def extend(self, blocks: list[Block]):
        self.blocks.extend(blocks)

    def insert(self, index, block: Block):
        self.blocks.insert(index, block)

    def to_dict(self):
        return {
            "role": self.role.value,
            "blocks": [block.to_dict() for block in self.blocks],
        }

    def __str__(self):
        return str(self.blocks)

    def __repr__(self):
        return f"Turn(blocks={self.blocks}, role={self.role})"


class Memory:
    """
    A class for storing the memory of a chat, organized by conversation turns.
    Each turn can contain multiple blocks (text, function calls, or structured data).
    """

    def __init__(self):
        # List of turns, where each turn is a list of blocks
        self.memory = []

    def new_turn(self, role: ROLE = ROLE.ASSISTANT):
        """Add a new conversation turn.

        Args:
            role (ROLE, optional): The role of the new turn. Defaults to ROLE.ASSISTANT.
        """
        self.memory.append(Turn([], role))

    def add_turn(
        self, blocks: list[Block] | list[FunctionCallBlock] | Block, role: ROLE
    ):
        """Add a new conversation turn containing one or more blocks.

        Args:
            blocks (list[Block] | Block): The blocks to add to the new turn.
            role (ROLE): The role of the new turn.
        """

        turn = Turn(blocks, role) if isinstance(blocks, list) else Turn([blocks], role)  # type: ignore

        self.memory.append(turn)

    def add_to_last_turn(self, block: Block):
        """Add a block to the most recent turn. Creates a new turn if memory is empty.
        Args:
            block (Block): The block to add to the most recent turn.
        """
        if not self.memory:
            self.memory.append(Turn([block], ROLE.ASSISTANT))
        else:
            self.memory[-1].append(block)

    def clear(self):
        """Clear all memory."""
        self.memory = []

    def __iter__(self):
        """Iterate through all blocks in all turns."""
        yield from self.memory

    def iter_blocks(self):
        """
        Iterate through blocks.
        """
        for turn in self.memory:
            yield from turn

    def copy(self):
        """Deep copy the memory."""
        from copy import deepcopy

        memory = Memory()
        memory.memory = deepcopy(self.memory)
        return memory

    def __len__(self):
        """Return the total number of turns."""
        return len(self.memory)

    def __getitem__(self, index):
        """Get all blocks from a specific turn."""
        return self.memory[index]

    def __setitem__(self, index, value):
        """Set blocks for a specific turn."""
        if isinstance(value, list):
            self.memory[index] = value
        else:
            self.memory[index] = [value]

    def __delitem__(self, index):
        """Delete a specific turn."""
        del self.memory[index]

    def __str__(self):
        """Return a string representation of the memory."""
        return str(self.memory)

    def __repr__(self):
        """Return a detailed string representation of the memory."""
        return f"Memory(turns={len(self)})"

    def __bool__(self):
        """Return True if memory contains any turns, False otherwise."""
        return bool(self.memory)

    def __eq__(self, other):
        """
        Compare two Memory objects based on their content hash.
        This is more efficient than comparing the full content structure.
        """
        if not isinstance(other, Memory):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        """
        Creates a deterministic hash based on the content of memory turns.
        """
        hash_components = []
        for turn in self.memory:
            turn_components = []
            for block in turn.blocks:
                turn_components.append(str(hash(block)))
            hash_components.append("|".join(turn_components))

        content_string = "||".join(hash_components)
        return int(hashlib.sha256(content_string.encode("utf-8")).hexdigest(), 16)

    def json_dumps(self) -> str:
        """Serialize the memory to JSON.

        Returns:
            str: The JSON representation of the memory.
        """
        return json.dumps(self.to_dict())

    def json_loads(self, json_str: str):
        """Deserialize JSON to the memory.

        Args:
            json_str (str): The JSON string to deserialize.
        """
        obj = json.loads(json_str)
        for t in obj:
            self.add_turn(
                blocks=[Block.from_dict(block) for block in t["blocks"]],
                role=ROLE(t["role"]),
            )

    def to_dict(self) -> list[dict]:
        """Convert memory to a dictionary.

        Returns:
            list[dict]: The dictionary representation of the memory.
        """
        return [turn.to_dict() for turn in self.memory]



import os
from collections.abc import AsyncIterator, Iterator
from typing import Any, Literal

import aioboto3
import boto3
from botocore.config import Config
from datapizza.core.cache import Cache
from datapizza.core.clients import Client, ClientResponse
from datapizza.core.clients.models import TokenUsage
from datapizza.memory import Memory
from datapizza.tools import Tool
from datapizza.type import FunctionCallBlock, TextBlock

from .memory_adapter import BedrockMemoryAdapter


class BedrockClient(Client):
    """A client for interacting with AWS Bedrock API.

    This class provides methods for invoking AWS Bedrock models (Claude, etc.)
    to generate responses. It extends the Client class and supports authentication
    via AWS profile or access keys.
    """

    def __init__(
        self,
        model: str = "anthropic.claude-3-5-sonnet-20241022-v2:0",
        system_prompt: str = "",
        temperature: float | None = None,
        cache: Cache | None = None,
        region_name: str = "us-east-1",
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        profile_name: str | None = None,
    ):
        """
        Args:
            model: The Bedrock model to use (e.g., 'anthropic.claude-3-5-sonnet-20241022-v2:0').
            system_prompt: The system prompt to use for the model.
            temperature: The temperature to use for generation (0-1 for most models).
            cache: The cache to use for responses.
            region_name: AWS region name (default: us-east-1).
            aws_access_key_id: AWS access key ID (optional, can use AWS_ACCESS_KEY_ID env var).
            aws_secret_access_key: AWS secret access key (optional, can use AWS_SECRET_ACCESS_KEY env var).
            aws_session_token: AWS session token (optional, for temporary credentials).
            profile_name: AWS profile name (optional, can use AWS_PROFILE env var).
        """
        if temperature and not 0 <= temperature <= 1:
            raise ValueError("Temperature must be between 0 and 1 for Bedrock models")

        super().__init__(
            model_name=model,
            system_prompt=system_prompt,
            temperature=temperature,
            cache=cache,
        )

        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id or os.getenv("AWS_ACCESS_KEY_ID")
        self.aws_secret_access_key = aws_secret_access_key or os.getenv(
            "AWS_SECRET_ACCESS_KEY"
        )
        self.aws_session_token = aws_session_token or os.getenv("AWS_SESSION_TOKEN")
        self.profile_name = profile_name or os.getenv("AWS_PROFILE")

        self.memory_adapter = BedrockMemoryAdapter()
        self._set_client()

    def _set_client(self):
        if not self.client:
            session_kwargs = {}

            # Priority: explicit credentials > profile > default credentials
            if self.aws_access_key_id and self.aws_secret_access_key:
                session_kwargs["aws_access_key_id"] = self.aws_access_key_id
                session_kwargs["aws_secret_access_key"] = self.aws_secret_access_key
                if self.aws_session_token:
                    session_kwargs["aws_session_token"] = self.aws_session_token
            elif self.profile_name:
                session_kwargs["profile_name"] = self.profile_name

            session = boto3.Session(**session_kwargs)

            # Create bedrock-runtime client with retry configuration
            config = Config(
                retries={"max_attempts": 3, "mode": "adaptive"},
                read_timeout=300,
            )

            self.client = session.client(
                service_name="bedrock-runtime",
                region_name=self.region_name,
                config=config,
            )

    def _set_a_client(self):
        """Initialize async bedrock-runtime client using aioboto3"""
        if not self.a_client:
            session_kwargs = {}

            # Priority: explicit credentials > profile > default credentials
            if self.aws_access_key_id and self.aws_secret_access_key:
                session_kwargs["aws_access_key_id"] = self.aws_access_key_id
                session_kwargs["aws_secret_access_key"] = self.aws_secret_access_key
                if self.aws_session_token:
                    session_kwargs["aws_session_token"] = self.aws_session_token
            elif self.profile_name:
                session_kwargs["profile_name"] = self.profile_name

            # Create async session
            session = aioboto3.Session(**session_kwargs)

            # Create bedrock-runtime client with retry configuration
            config = Config(
                retries={"max_attempts": 3, "mode": "adaptive"},
                read_timeout=300,
            )

            # Store the session and config for async client creation
            self.a_session = session
            self.a_config = config
            self.a_region_name = self.region_name

    def _convert_tools(self, tools: list[Tool]) -> list[dict[str, Any]]:
        """Convert tools to Bedrock tool format (similar to Anthropic)"""
        bedrock_tools = []
        for tool in tools:
            bedrock_tool = {
                "toolSpec": {
                    "name": tool.name,
                    "description": tool.description or "",
                    "inputSchema": {
                        "json": {
                            "type": "object",
                            "properties": tool.properties,
                            "required": tool.required,
                        }
                    },
                }
            }
            bedrock_tools.append(bedrock_tool)
        return bedrock_tools

    def _convert_tool_choice(
        self, tool_choice: Literal["auto", "required", "none"] | list[str]
    ) -> dict:
        """Convert tool choice to Bedrock format"""
        if isinstance(tool_choice, list):
            if len(tool_choice) > 1:
                raise NotImplementedError(
                    "Multiple function names not supported by Bedrock"
                )
            return {"tool": {"name": tool_choice[0]}}
        elif tool_choice == "required":
            return {"any": {}}
        elif tool_choice == "auto":
            return {"auto": {}}
        else:  # none
            return {}

    def _response_to_client_response(
        self, response: dict, tool_map: dict[str, Tool] | None = None
    ) -> ClientResponse:
        """Convert Bedrock response to ClientResponse"""
        blocks = []

        # Parse the response body
        response_body = response.get("output", {})

        # Handle message content
        message = response_body.get("message", {})
        content_items = message.get("content", [])

        for content_item in content_items:
            if "text" in content_item:
                blocks.append(TextBlock(content=content_item["text"]))
            elif "toolUse" in content_item:
                tool_use = content_item["toolUse"]
                tool = tool_map.get(tool_use["name"]) if tool_map else None
                if not tool:
                    raise ValueError(f"Tool {tool_use['name']} not found")

                blocks.append(
                    FunctionCallBlock(
                        id=tool_use["toolUseId"],
                        name=tool_use["name"],
                        arguments=tool_use["input"],
                        tool=tool,
                    )
                )

        # Extract usage information
        usage = response.get("usage", {})
        prompt_tokens = usage.get("inputTokens", 0)
        completion_tokens = usage.get("outputTokens", 0)

        # Extract stop reason
        stop_reason = response.get("stopReason")

        return ClientResponse(
            content=blocks,
            stop_reason=stop_reason,
            usage=TokenUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                cached_tokens=0,  # Bedrock doesn't expose cache metrics in the same way
            ),
        )

    def _invoke(
        self,
        *,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        **kwargs,
    ) -> ClientResponse:
        """Implementation of the abstract _invoke method for Bedrock"""
        if tools is None:
            tools = []

        client = self._get_client()
        messages = self._memory_to_contents(None, input, memory)

        # Remove model role messages (Bedrock doesn't support this)
        messages = [message for message in messages if message.get("role") != "model"]

        tool_map = {tool.name: tool for tool in tools}

        # Build the request body according to Bedrock Converse API
        request_body = {
            "modelId": self.model_name,
            "messages": messages,
            "inferenceConfig": {
                "maxTokens": max_tokens or 2048,
            },
        }

        if temperature is not None:
            request_body["inferenceConfig"]["temperature"] = temperature

        # Add system prompt if provided
        if system_prompt:
            request_body["system"] = [{"text": system_prompt}]

        # Add tools if provided
        if tools:
            request_body["toolConfig"] = {
                "tools": self._convert_tools(tools),
                "toolChoice": self._convert_tool_choice(tool_choice),
            }

        # Add any additional kwargs
        request_body.update(kwargs)

        # Call Bedrock Converse API
        response = client.converse(**request_body)

        return self._response_to_client_response(response, tool_map)

    async def _a_invoke(
        self,
        *,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        **kwargs,
    ) -> ClientResponse:
        """Async implementation using aioboto3"""
        if tools is None:
            tools = []

        # Ensure async client is initialized
        if not hasattr(self, "a_session"):
            self._set_a_client()

        messages = self._memory_to_contents(None, input, memory)

        # Remove model role messages (Bedrock doesn't support this)
        messages = [message for message in messages if message.get("role") != "model"]

        tool_map = {tool.name: tool for tool in tools}

        # Build the request body according to Bedrock Converse API
        request_body = {
            "modelId": self.model_name,
            "messages": messages,
            "inferenceConfig": {
                "maxTokens": max_tokens or 2048,
            },
        }

        if temperature is not None:
            request_body["inferenceConfig"]["temperature"] = temperature

        # Add system prompt if provided
        if system_prompt:
            request_body["system"] = [{"text": system_prompt}]

        # Add tools if provided
        if tools:
            request_body["toolConfig"] = {
                "tools": self._convert_tools(tools),
                "toolChoice": self._convert_tool_choice(tool_choice),
            }

        # Add any additional kwargs
        request_body.update(kwargs)

        # Call Bedrock Converse API asynchronously
        async with self.a_session.client(
            service_name="bedrock-runtime",
            region_name=self.a_region_name,
            config=self.a_config,
        ) as client:
            response = await client.converse(**request_body)

        return self._response_to_client_response(response, tool_map)

    def _stream_invoke(
        self,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        **kwargs,
    ) -> Iterator[ClientResponse]:
        """Implementation of streaming for Bedrock"""
        if tools is None:
            tools = []

        client = self._get_client()
        messages = self._memory_to_contents(None, input, memory)

        # Remove model role messages
        messages = [message for message in messages if message.get("role") != "model"]

        # Build the request body
        request_body = {
            "modelId": self.model_name,
            "messages": messages,
            "inferenceConfig": {
                "maxTokens": max_tokens or 2048,
            },
        }

        if temperature is not None:
            request_body["inferenceConfig"]["temperature"] = temperature

        if system_prompt:
            request_body["system"] = [{"text": system_prompt}]

        if tools:
            request_body["toolConfig"] = {
                "tools": self._convert_tools(tools),
                "toolChoice": self._convert_tool_choice(tool_choice),
            }

        request_body.update(kwargs)

        # Call Bedrock ConverseStream API
        response = client.converse_stream(**request_body)

        # Process streaming response
        message_text = ""
        input_tokens = 0
        output_tokens = 0
        stop_reason = None

        stream = response.get("stream")
        if stream:
            for event in stream:
                if "contentBlockDelta" in event:
                    delta = event["contentBlockDelta"].get("delta", {})
                    if "text" in delta:
                        text_delta = delta["text"]
                        message_text += text_delta
                        yield ClientResponse(
                            content=[TextBlock(content=message_text)],
                            delta=text_delta,
                            stop_reason=None,
                        )

                elif "metadata" in event:
                    metadata = event["metadata"]
                    usage = metadata.get("usage", {})
                    input_tokens = usage.get("inputTokens", 0)
                    output_tokens = usage.get("outputTokens", 0)

                elif "messageStop" in event:
                    stop_reason = event["messageStop"].get("stopReason")

        # Final response with complete information
        yield ClientResponse(
            content=[TextBlock(content=message_text)],
            delta="",
            stop_reason=stop_reason,
            usage=TokenUsage(
                prompt_tokens=input_tokens,
                completion_tokens=output_tokens,
                cached_tokens=0,
            ),
        )

    async def _a_stream_invoke(
        self,
        input: str,
        tools: list[Tool] | None,
        memory: Memory | None,
        tool_choice: Literal["auto", "required", "none"] | list[str],
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        **kwargs,
    ) -> AsyncIterator[ClientResponse]:
        """Async streaming implementation using aioboto3"""
        if tools is None:
            tools = []

        # Ensure async client is initialized
        if not hasattr(self, "a_session"):
            self._set_a_client()

        messages = self._memory_to_contents(None, input, memory)

        # Remove model role messages
        messages = [message for message in messages if message.get("role") != "model"]

        # Build the request body
        request_body = {
            "modelId": self.model_name,
            "messages": messages,
            "inferenceConfig": {
                "maxTokens": max_tokens or 2048,
            },
        }

        if temperature is not None:
            request_body["inferenceConfig"]["temperature"] = temperature

        if system_prompt:
            request_body["system"] = [{"text": system_prompt}]

        if tools:
            request_body["toolConfig"] = {
                "tools": self._convert_tools(tools),
                "toolChoice": self._convert_tool_choice(tool_choice),
            }

        request_body.update(kwargs)

        # Call Bedrock ConverseStream API asynchronously
        async with self.a_session.client(
            service_name="bedrock-runtime",
            region_name=self.a_region_name,
            config=self.a_config,
        ) as client:
            response = await client.converse_stream(**request_body)

            # Process streaming response
            message_text = ""
            input_tokens = 0
            output_tokens = 0
            stop_reason = None

            stream = response.get("stream")
            if stream:
                async for event in stream:
                    if "contentBlockDelta" in event:
                        delta = event["contentBlockDelta"].get("delta", {})
                        if "text" in delta:
                            text_delta = delta["text"]
                            message_text += text_delta
                            yield ClientResponse(
                                content=[TextBlock(content=message_text)],
                                delta=text_delta,
                                stop_reason=None,
                            )

                    elif "metadata" in event:
                        metadata = event["metadata"]
                        usage = metadata.get("usage", {})
                        input_tokens = usage.get("inputTokens", 0)
                        output_tokens = usage.get("outputTokens", 0)

                    elif "messageStop" in event:
                        stop_reason = event["messageStop"].get("stopReason")

            # Final response with complete information
            yield ClientResponse(
                content=[TextBlock(content=message_text)],
                delta="",
                stop_reason=stop_reason,
                usage=TokenUsage(
                    prompt_tokens=input_tokens,
                    completion_tokens=output_tokens,
                    cached_tokens=0,
                ),
            )

    def _structured_response(
        self,
        input: str,
        output_cls: type,
        memory: Memory | None,
        temperature: float | None,
        max_tokens: int,
        system_prompt: str | None,
        tools: list[Tool] | None,
        tool_choice: Literal["auto", "required", "none"] | list[str] = "auto",
        **kwargs,
    ) -> ClientResponse:
        """Bedrock doesn't natively support structured output like OpenAI

        This would require prompting techniques or using Anthropic's prompt caching
        """
        raise NotImplementedError(
            "Bedrock doesn't natively support structured responses. "
            "Consider using prompt engineering or JSON mode with validation."
        )

    async def _a_structured_response(self, *args, **kwargs):
        raise NotImplementedError(
            "Bedrock doesn't natively support structured responses"
        )



import logging
from copy import deepcopy
from dataclasses import dataclass

from datapizza.core.models import ChainableProducer, PipelineComponent
from datapizza.core.utils import replace_env_vars

log = logging.getLogger(__name__)


@dataclass
class Edge:
    from_node_name: str
    to_node_name: str
    src_key: str | None
    dst_key: str


class DagPipeline:
    """
    A pipeline that runs a graph of a dependency graph.
    """

    def __init__(self):
        self.nodes: dict[str, PipelineComponent] = {}
        self.edges = []

    # get the nodes that depend on the given node
    def _get_edges_from(self, node_name: str) -> list[Edge]:
        return [d for d in self.edges if d.from_node_name == node_name]

    # get the nodes that the given node depends on
    def _get_edges_to(self, node_name: str) -> list[Edge]:
        return [d for d in self.edges if d.to_node_name == node_name]

    def add_module(self, node_name: str, node: PipelineComponent):
        """
        Add a module to the pipeline.

        Args:
            node_name (str): The name of the module.
            node (PipelineComponent): The module to add.
        """
        # Nodes must be ChainableProducer or PipelineComponent or callable
        if isinstance(node, ChainableProducer):
            module_component = node.as_module_component()
            self.nodes[node_name] = module_component

        elif isinstance(node, PipelineComponent) or callable(node):
            self.nodes[node_name] = node
        else:
            raise ValueError(
                f"Node {node_name} must be a ChainableProducer, PipelineComponent, or callable."
            )

    def connect(
        self,
        source_node: str,
        target_node: str,
        target_key: str,
        source_key: str | None = None,
    ):
        """
        Connect two nodes in the pipeline.

        Args:
            source_node (str): The name of the source node.
            target_node (str): The name of the target node.
            target_key (str): The key to store the result of the target node in the source node.
            source_key (str, optional): The key to retrieve the result of the source node from the target node. Defaults to None.
        """
        self.edges.append(
            Edge(
                from_node_name=source_node,
                to_node_name=target_node,
                src_key=source_key,
                dst_key=target_key,
            )
        )

    def _get_nodes_ready_to_run(self, results: dict):
        ready_nodes = []
        for node_name in self.nodes:
            # continue if node is already processed
            if node_name in results:
                continue

            # If node has no dependencies, it's ready to be processed
            if not self._get_edges_to(node_name):
                ready_nodes.append(node_name)
                continue

            # Check if all dependencies are resolved
            dependencies = [d.from_node_name for d in self._get_edges_to(node_name)]
            if all(dep in results for dep in dependencies):
                ready_nodes.append(node_name)

        return ready_nodes

    def _get_args_for_node(self, node_name: str, _input: dict, results: dict) -> dict:
        # Get all the nodes that must be run before the given node
        # get the results from the previous nodes and return them as a dict

        previous_nodes = self._get_edges_to(node_name)
        # Start with the original input
        args = deepcopy(_input.get(node_name, {}))

        for edge in previous_nodes:
            from_node_result = results[edge.from_node_name]

            # Extract the correct value from source
            if edge.src_key:
                value = from_node_result.get(edge.src_key)
                if value is None and edge.src_key not in from_node_result:
                    log.warning(
                        f"Source key '{edge.src_key}' not found in result of node '{edge.from_node_name}'. Using None."
                    )

            else:
                value = from_node_result

            # Place it in the right location in args
            if edge.dst_key:
                args[edge.dst_key] = value  # TODO  deepcopy(value)
            else:
                raise ValueError(
                    f"No destination key provided for node '{node_name}' from '{edge.from_node_name}'."
                )

        return args

    def run(self, data: dict) -> dict:
        """
        Run the pipeline.

        Args:
            data (dict): The input data to the pipeline.

        Returns:
            dict: The results of the pipeline.
        """
        processed_nodes = set()
        pipeline_results = {}
        queue = self._get_nodes_ready_to_run(pipeline_results)

        while queue:
            node_name = queue.pop(0)

            if node_name in processed_nodes:
                continue

            node = self.nodes[node_name]
            try:
                arguments = self._get_args_for_node(node_name, data, pipeline_results)
                log.debug(f"Arguments for node {node_name}: {list(arguments.keys())}")
                node_result = node(**arguments)
                pipeline_results[node_name] = node_result
                processed_nodes.add(node_name)

                newly_ready = self._get_nodes_ready_to_run(pipeline_results)
                for ready_node in newly_ready:
                    if ready_node not in processed_nodes and ready_node not in queue:
                        queue.append(ready_node)

            except Exception as e:
                log.error(f"Error running node {node_name}: {e!s}")
                raise

        if len(processed_nodes) < len(self.nodes):
            unprocessed = set(self.nodes.keys()) - processed_nodes
            log.warning(
                f"Not all nodes were processed. Unprocessed: {unprocessed}. Possible cycle or error."
            )
            # This could indicate a cycle in the graph or an earlier error

        return pipeline_results

    async def a_run(self, data: dict):
        """
        Run the pipeline asynchronously.

        Args:
            data (dict): The input data to the pipeline.

        Returns:
            dict: The results of the pipeline.
        """
        processed_nodes = set()
        pipeline_results = {}
        queue = self._get_nodes_ready_to_run(pipeline_results)

        while queue:
            node_name = queue.pop(0)

            if node_name in processed_nodes:
                continue

            node = self.nodes[node_name]
            try:
                arguments = self._get_args_for_node(node_name, data, pipeline_results)
                log.debug(f"Arguments for node {node_name}: {list(arguments.keys())}")
                node_result = await node.a_run(**arguments)
                pipeline_results[node_name] = node_result
                processed_nodes.add(node_name)

                newly_ready = self._get_nodes_ready_to_run(pipeline_results)
                for ready_node in newly_ready:
                    if ready_node not in processed_nodes and ready_node not in queue:
                        queue.append(ready_node)

            except Exception as e:
                log.error(f"Error running node {node_name}: {e!s}")
                raise

        if len(processed_nodes) < len(self.nodes):
            unprocessed = set(self.nodes.keys()) - processed_nodes
            log.warning(
                f"Not all nodes were processed. Unprocessed: {unprocessed}. Possible cycle or error."
            )

        return pipeline_results

    def from_yaml(self, config_path: str) -> "DagPipeline":
        """
        Load the pipeline from a YAML configuration file.

        Args:
            config_path (str): Path to the YAML configuration file.

        Returns:
            DagPipeline: The pipeline instance.
        """
        import importlib

        import yaml

        from datapizza.clients import ClientFactory

        with open(config_path) as file:
            config = yaml.safe_load(file)

        constants = config.get("constants", [])
        config = replace_env_vars(config, constants)

        dag_pipeline = config["dag_pipeline"]
        clients = {}

        if "clients" in dag_pipeline:
            for client_name, client_config in dag_pipeline["clients"].items():
                provider = client_config.pop("provider")
                client = ClientFactory.create(provider, **client_config)
                clients[client_name] = client

        self.nodes = {}
        self.edges = []

        if "modules" in dag_pipeline:
            for module_config in dag_pipeline["modules"]:
                try:
                    module_name = module_config["name"]
                    module_path = module_config["module"]
                    module_import = importlib.import_module(module_path)
                    class_ = getattr(module_import, module_config["type"])

                    params = module_config.get("params", {})

                    if "client" in params:
                        client_name = params["client"]
                        if client_name not in clients:
                            raise ValueError(
                                f"Client '{client_name}' not found in clients configuration for node {module_name}"
                            )
                        params["client"] = clients[client_name]

                    component_instance = class_(**params)
                    self.add_module(module_name, component_instance)
                except (ImportError, AttributeError) as e:
                    raise ValueError(
                        f"Could not load module {module_config.get('type', 'N/A')} for node {module_config.get('name', 'N/A')}: {e!s}"
                    ) from e
                except KeyError as e:
                    raise ValueError(
                        f"Missing required key {e!s} in module configuration: {module_config}"
                    ) from e

        if "connections" in dag_pipeline:
            for connection in dag_pipeline["connections"]:
                from_node = connection["from"]
                to_node = connection["to"]
                if from_node not in self.nodes:
                    raise ValueError(
                        f"Source node '{from_node}' for connection not found in loaded modules."
                    )
                if to_node not in self.nodes:
                    raise ValueError(
                        f"Target node '{to_node}' for connection not found in loaded modules."
                    )

                source_key = connection.get("source_key")
                target_key = connection.get("target_key")
                self.connect(
                    from_node,
                    to_node,
                    target_key=target_key,
                    source_key=source_key,
                )

        return self



from datapizza.core.models import PipelineComponent
from datapizza.pipeline import Dependency, FunctionalPipeline


def test_functional_pipeline():
    class A(PipelineComponent):
        def _run(self):
            return "A"

        async def _a_run(self):
            return "A"

    class B(PipelineComponent):
        def _run(self):
            return "B"

        async def _a_run(self):
            return "B"

    pipeline = FunctionalPipeline()
    pipeline.run("start", A())
    pipeline.run("then", B())
    assert pipeline.execute() == {"start": "A", "then": "B"}


def test_functional_pipeline_with_dependencies():
    class A(PipelineComponent):
        def _run(self):
            return "A"

        async def _a_run(self):
            return "A"

    class B(PipelineComponent):
        def _run(self, data):
            return data + "B"

        async def _a_run(self, data):
            return data + "B"

    pipeline = FunctionalPipeline()
    pipeline.run("start", A())
    pipeline.run(
        "then", B(), dependencies=[Dependency(node_name="start", target_key="data")]
    )
    assert pipeline.execute() == {"start": "A", "then": "AB"}


def test_pipeline_with_then():
    class A(PipelineComponent):
        def _run(self):
            return "A"

        async def _a_run(self):
            return "A"

    class B(PipelineComponent):
        def _run(self, data):
            return "B"

        async def _a_run(self, data):
            return "B"

    pipeline = FunctionalPipeline()
    pipeline.run("start", A())
    pipeline.then("then", B(), target_key="data")
    assert pipeline.execute() == {"start": "A", "then": "B"}


def test_pipeline_with_then_and_dependencies():
    class A(PipelineComponent):
        def _run(self):
            return "A"

        async def _a_run(self):
            return "A"

    class C(PipelineComponent):
        def _run(self):
            return "C"

        async def _a_run(self):
            return "C"

    class B(PipelineComponent):
        def _run(self, data, c):
            return data + "B" + c

        async def _a_run(self, data, c):
            return data + "B" + c

    pipeline = FunctionalPipeline()
    pipeline.run("C", C())  # NOW C MUST BE RUN FIRST of B. In the future, we can
    # make this more flexible
    pipeline.run("start", A())
    pipeline.then(
        "then",
        B(),
        target_key="data",
        dependencies=[Dependency(node_name="C", target_key="c")],
    )
    assert pipeline.execute() == {"start": "A", "then": "ABC", "C": "C"}


def test_pipeline_with_get():
    class A(PipelineComponent):
        def _run(self):
            return "A"

        async def _a_run(self):
            return "A"

    pipeline = FunctionalPipeline().run("start", A()).get("start")
    res = pipeline.execute()

    assert res == "A"


def test_pipeline_from_yaml():
    from datapizza.modules.rewriters import ToolRewriter

    pipeline = FunctionalPipeline.from_yaml(
        "datapizza-ai-core/datapizza/pipeline/tests/functional_pipeline_config.yaml"
    )
    assert pipeline is not None
    assert pipeline.nodes[0].get("name") == "rewriter"
    assert pipeline.nodes[0].get("node").__class__ == ToolRewriter
    assert pipeline.nodes[0].get("node").client.model_name == "gpt-4o"


def test_pipeline_with_foreach():
    class A(PipelineComponent):
        def _run(self):
            return [1, 2, 3]

        async def _a_run(self):
            return [1, 2, 3]

    class B(PipelineComponent):
        def _run(self, data):
            return str(data) + "B"

        async def _a_run(self, data):
            return str(data) + "B"

    pipeline = (
        FunctionalPipeline()
        .run("start", A())
        .foreach(
            "foreach",
            dependencies=[Dependency(node_name="start")],
            do=B(),
        )
    )

    assert pipeline.execute() == {
        "start": [1, 2, 3],
        "foreach": ["1B", "2B", "3B"],
    }


def test_pipeline_with_branch():
    class A(PipelineComponent):
        def _run(self):
            return "A"

        async def _a_run(self):
            return "A"

    class B(PipelineComponent):
        def _run(self):
            return "B"

        async def _a_run(self):
            return "B"

    class C(PipelineComponent):
        def _run(self):
            return "C"

        async def _a_run(self):
            return "C"

    # Test branch with TRUE condition
    pipeline = (
        FunctionalPipeline()
        .run("start", A())
        .branch(
            condition=lambda context: context["start"] == "A",
            if_true=FunctionalPipeline().run("B", B()),
            if_false=FunctionalPipeline().run("C", C()),
        )
    )
    assert pipeline.execute() == {"start": "A", "B": "B"}

    # Test branch with FALSE condition
    pipeline = (
        FunctionalPipeline()
        .run("start", A())
        .branch(
            condition=lambda context: context["start"] == "B",
            if_true=FunctionalPipeline().run("B", B()),
            if_false=FunctionalPipeline().run("C", C()),
        )
    )
    assert pipeline.execute() == {"start": "A", "C": "C"}



from abc import abstractmethod
from typing import overload

from datapizza.core.models import PipelineComponent
from datapizza.type.type import Chunk, Node


class Splitter(PipelineComponent):
    def _run(self, node: Node) -> list[Chunk]:
        return self.split(node)

    async def _a_run(self, node: Node) -> list[Chunk]:
        return await self.a_split(node)

    @abstractmethod
    @overload
    def split(self, text: str) -> list[Chunk]:
        pass

    @abstractmethod
    @overload
    def split(self, node: Node) -> list[Chunk]:
        pass

    async def a_split(self, text: str | Node) -> list[Chunk]:
        raise NotImplementedError(
            f"a_split is not implemented in {self.__class__.__name__} "
        )



import asyncio
from concurrent.futures import Future, ThreadPoolExecutor

from datapizza.core.clients import Client
from datapizza.core.modules.captioner import NodeCaptioner
from datapizza.type import Media, MediaBlock, MediaNode, Node, NodeType


class LLMCaptioner(NodeCaptioner):
    """
    Captioner that uses an LLM client to caption a node.
    """

    def __init__(
        self,
        client: Client,
        max_workers: int = 3,
        system_prompt_table: str | None = "Generate concise captions for tables.",
        system_prompt_figure: str | None = "Generate descriptive captions for figures.",
    ):
        """
        Captioner that uses an LLM client to caption a node.
        Args:
            client: The LLM client to use.
            max_workers: The maximum number of workers to use. in sync mode is the number of threads spawned, in async mode is the number of batches.
            system_prompt_table: The system prompt to use for table captioning.
            system_prompt_figure: The system prompt to use for figure captioning.
        """
        self.client = client
        self.max_workers = max_workers
        self.system_prompt_figure = system_prompt_figure
        self.system_prompt_table = system_prompt_table

    def _get_all_media_nodes(self, node: Node) -> list[MediaNode]:
        if isinstance(node, MediaNode):
            return [node]

        media_nodes = []
        for child in node.children:
            media_nodes.extend(self._get_all_media_nodes(child))
        return media_nodes

    def _process_media(self, media_node: MediaNode) -> tuple[MediaNode, str]:
        if media_node.node_type == NodeType.FIGURE:
            system_prompt = self.system_prompt_figure
        elif media_node.node_type == NodeType.TABLE:
            system_prompt = self.system_prompt_table
        else:
            raise ValueError(f"Unsupported node type: {media_node.node_type}")

        caption = self.caption_media(
            media=media_node.media, system_prompt=system_prompt
        )

        # Create a new MediaNode with the caption nodes
        new_children = [
            Node(
                node_type=NodeType.PARAGRAPH,
                content=f" {media_node.content} <{media_node.node_type.value}> [{caption}]",
                metadata=media_node.metadata,
            ),
            *media_node.children,
            Node(
                node_type=NodeType.PARAGRAPH,
                content=f"</{media_node.node_type.value}>",
                metadata=media_node.metadata,
            ),
        ]

        new_media_node = MediaNode(
            node_type=media_node.node_type,
            media=media_node.media,
            children=new_children,
            metadata=media_node.metadata,
        )

        return new_media_node, caption

    def _replace_media_nodes(
        self, node: Node, processed_nodes: list[MediaNode]
    ) -> Node:
        if isinstance(node, MediaNode):
            # Find matching processed node
            for processed_node in processed_nodes:
                if processed_node.media == node.media:
                    return processed_node
            return node

        # Create new node with processed children
        new_children = [
            self._replace_media_nodes(child, processed_nodes) for child in node.children
        ]
        return Node(
            node_type=node.node_type,
            content=node.content,
            children=new_children,
            metadata=node.metadata,
        )

    def caption(self, node: Node) -> Node:
        """
        Caption a node.
        Args:
            node: The node to caption.

        Returns:
            The same node with the caption.
        """
        media_nodes = self._get_all_media_nodes(node)

        # Process media nodes concurrently
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures: list[Future] = [
                executor.submit(self._process_media, media) for media in media_nodes
            ]

            # Wait for all futures to complete and collect results
            processed_nodes = []
            for future in futures:
                processed_node, _ = future.result()
                processed_nodes.append(processed_node)

        return self._replace_media_nodes(node, processed_nodes)

    def caption_media(self, media: Media, system_prompt: str | None = None) -> str:
        """
        Caption an image.
        Args:
            media: The media to caption.
            system_prompt: Optional system prompt to guide the captioning.

        Returns:
            The string caption.
        """
        response = self.client.invoke(
            input=[MediaBlock(media)], system_prompt=system_prompt
        )
        return response.text

    async def a_caption(self, node: Node) -> Node:
        """
        async Caption a node.
        Args:
            node: The node to caption.

        Returns:
            The same node with the caption.
        """
        media_nodes = self._get_all_media_nodes(node)

        # Process media nodes in batches based on max_workers
        processed_nodes = []
        for i in range(0, len(media_nodes), self.max_workers):
            batch = media_nodes[i : i + self.max_workers]

            # Process batch concurrently
            results = await asyncio.gather(
                *[self._a_process_media(media) for media in batch]
            )

            # Unpack results from this batch
            processed_nodes.extend([node for node, _ in results])

        return self._replace_media_nodes(node, processed_nodes)

    async def a_caption_media(
        self, media: Media, system_prompt: str | None = None
    ) -> str:
        """
        async Caption image.
        Args:
            media: The media to caption.
            system_prompt: Optional system prompt to guide the captioning.

        Returns:
            The string caption.
        """
        response = await self.client.a_invoke(
            input=[MediaBlock(media)], system_prompt=system_prompt
        )
        return response.text

    async def _a_process_media(self, media_node: MediaNode) -> tuple[MediaNode, str]:
        if media_node.node_type == NodeType.FIGURE:
            system_prompt = self.system_prompt_figure
        elif media_node.node_type == NodeType.TABLE:
            system_prompt = self.system_prompt_table
        else:
            raise ValueError(f"Unsupported node type: {media_node.node_type}")

        caption = await self.a_caption_media(
            media=media_node.media, system_prompt=system_prompt
        )

        # Create a new MediaNode with the caption nodes
        new_children = [
            Node(
                node_type=NodeType.PARAGRAPH,
                content=f" {media_node.content} <{media_node.node_type.value}> [{caption}]",
                metadata=media_node.metadata,
            ),
            *media_node.children,
            Node(
                node_type=NodeType.PARAGRAPH,
                content=f"</{media_node.node_type.value}>",
                metadata=media_node.metadata,
            ),
        ]

        new_media_node = MediaNode(
            node_type=media_node.node_type,
            media=media_node.media,
            children=new_children,
            metadata=media_node.metadata,
        )

        return new_media_node, caption



import hashlib
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import wraps

log = logging.getLogger(__name__)


class Cache(ABC):
    """
    This is the abstract base class for all cache implementations.
    Concrete subclasses must provide implementations for the abstract methods that define how caching is handled.


    When a cache instance is attached to a client, it will automatically store the results of the client`s method calls.
    If the same method is invoked multiple times with identical arguments, the cache returns the stored result instead of re-executing the method.
    """

    @abstractmethod
    def get(self, key: str) -> object:
        """
        Retrieve an object from the cache.

        Args:
            key (str): The key to retrieve the object for.

        Returns:
            The object stored in the cache.
        """

    @abstractmethod
    def set(self, key: str, value: str):
        """
        Store an object in the cache.

        Args:
            key (str): The key to store the object for.
            value (str): The object to store in the cache.
        """


def cacheable(key_func: Callable):
    """
    Decorator that caches function results based on key functions.

    Args:
        *key_funcs: Functions that take the function's self and arguments and return a value for the cache key
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.cache:
                return func(self, *args, **kwargs)

            try:
                # Get all arguments as a dictionary
                bound_args = kwargs.copy()
                func_args = func.__code__.co_varnames[: func.__code__.co_argcount]
                for i, arg_name in enumerate(func_args[1:]):  # Skip 'self'
                    if i < len(args):
                        bound_args[arg_name] = args[i]

                # Generate cache key using the provided functions

                cache_key = key_func(self, bound_args)
                # hash the string
                cache_key = hashlib.sha256(cache_key.encode()).hexdigest()

                cached_result = self.cache.get(cache_key)
                if cached_result is not None:
                    log.info(f"Cache hit for {cache_key}")
                    return cached_result

            except Exception as e:
                log.warning(f"Error generating cache key for {func.__name__}: {e}")
                return func(self, *args, **kwargs)

            # Execute function if not cached
            result = func(self, *args, **kwargs)
            try:
                self.cache.set(cache_key, result)
            except Exception as e:
                log.error(f"Error setting cache for {cache_key}: {e}")
            return result

        return wrapper

    return decorator


class MemoryCache(Cache):
    """
    A simple in-memory cache implementation.
    """

    def __init__(self):
        self.cache = {}

    def get(self, key: str) -> object:
        """Retrieve an object from the cache."""
        return self.cache.get(key)

    def set(self, key: str, value: str):
        """Set an object in the cache."""
        self.cache[key] = value



from abc import abstractmethod

from datapizza.core.clients import Client
from datapizza.core.models import PipelineComponent
from datapizza.type.type import Chunk


class Metatagger(PipelineComponent):
    """
    A meta tagger that can tag a node.
    """

    def __init__(self, client: Client):
        self.client = client

    @abstractmethod
    def tag(self, chunks: list[Chunk]):
        pass

    def a_tag(self, chunks: list[Chunk]):
        raise NotImplementedError

    def _run(self, chunks: list[Chunk]):
        return self.tag(chunks)

    async def _a_run(self, chunk: list[Chunk]):
        return await self.a_tag(chunk)



from threading import Lock

from datapizza.core.clients import Client


class ClientManager:
    _instance: Client | None = None
    _lock = Lock()

    @classmethod
    def set_global_client(cls, client: Client) -> None:
        """Set the global Client instance.

        Args:
            config: Client instance to be used globally
        """
        with cls._lock:
            cls._instance = client

    @classmethod
    def get_global_client(cls) -> Client | None:
        """Get the current global Client instance.

        Returns:
            The global client instance if set, None otherwise
        """
        return cls._instance

    @classmethod
    def clear_global_client(cls) -> None:
        """Clear the global Client instance."""
        with cls._lock:
            cls._instance = None



import fnmatch
import os
import re

from datapizza.tools import tool


def string_matches_patterns(string_to_check: str, patterns: list[str]) -> bool:
    def _check_pattern(string_to_check: str, pattern: str) -> bool:
        # glob patterns
        regex = fnmatch.translate(pattern.lower())
        if re.match(regex, string_to_check.lower()):
            return True
        # regex patterns
        try:
            if re.match(pattern, string_to_check.lower()):
                return True
        except re.error:
            pass

        return False

    if len(patterns) == 0:
        return True

    return any(_check_pattern(string_to_check, pattern) for pattern in patterns)


class FileSystem:
    """A collection of tools for interacting with the local file system."""

    def __init__(self, paths_to_include=None, paths_to_exclude=None) -> None:
        """
        Initialize the FileSystem. You can set `paths_to_include` and `paths_to_exclude` as glob or regex patterns on paths to reduce the scope of the tool.
        By default, all paths are included. Exclusion patterns are evaluated after the inclusion patterns, Therefore, exclusion patterns should apply subfilters on the inclusion patterns to be effective.
        Example usage::

                        FileSystem() #includes the whole file system
                        FileSystem(paths_to_include=["/project/dir/*"]) # includes all the files in /project/dir/
                        FileSystem(paths_to_exclude=["/project/dir/.env"], paths_to_include=["/project/dir/*"]) # includes all the files and directories in /project/dir/ except .env
                        FileSystem(paths_to_exclude=["*/.env"], paths_to_include=["/project/dir/*"]) # as above, includes all the files and directories in /project/dir/ except .env
                        FileSystem(paths_to_exclude=["/data/archive.zip"], paths_to_include=["*.zip", "*.txt"]) #includes all txt and zip files except /data/archive.zip

        Args:
                paths_to_include (list[str], optional): Define a list of glob or regular expression patterns for paths to include. Defaults to None.
                paths_to_exclude (list[str], optional): Define a list of glob or regular expression patterns to exclude. It is evaluated after inclusion patterns. Defaults to None.
        """
        self.include_patterns = paths_to_include if paths_to_include else ["*"]
        self.exclude_patterns = paths_to_exclude if paths_to_exclude else []

    def is_path_valid(self, path: str) -> bool:
        if string_matches_patterns(path, self.include_patterns):
            return not (
                self.exclude_patterns
                and string_matches_patterns(path, self.exclude_patterns)
            )
        return False

    @tool
    def list_directory(self, path: str) -> str:
        """
        Lists all valid files and directories in a given path.
        :param path: The path of the directory to list.
        """

        if not os.path.isdir(path):
            return f"Error: Path '{path}' is not a valid directory."

        try:
            entries = os.listdir(path)
            if not entries:
                return f"The directory '{path}' is empty."

            formatted_entries = []
            for entry in entries:
                entry_path = os.path.join(path, entry)
                if self.is_path_valid(entry_path):
                    if os.path.isdir(entry_path):
                        formatted_entries.append(f"[DIR] {entry}")
                    else:
                        formatted_entries.append(f"[FILE] {entry}")

            return "\n".join(formatted_entries)
        except Exception as e:
            return f"An error occurred: {e}"

    @tool
    def read_file(self, file_path: str) -> str:
        """
        Reads the content of a specified file.
        :param file_path: The path of the file to read.
        """
        if not self.is_path_valid(file_path):
            return f"Path '{file_path}' is outside the tool's scope."
        try:
            with open(file_path, encoding="utf-8") as f:
                return f.read()
        except FileNotFoundError:
            return f"Error: File '{file_path}' not found."
        except Exception as e:
            return f"An error occurred: {e}"

    @tool
    def write_file(self, file_path: str, content: str) -> str:
        """
        Writes content to a specified file. Creates the file if it does not exist.
        :param file_path: The path of the file to write to.
        :param content: The content to write to the file.
        """
        if not self.is_path_valid(file_path):
            return f"Path '{file_path}' is outside the tool's scope."
        try:
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(content)
            return f"Successfully wrote to file '{file_path}'."
        except Exception as e:
            return f"An error occurred: {e}"

    @tool
    def create_directory(self, path: str) -> str:
        """
        Creates a new directory at the specified path.
        :param path: The path where the new directory should be created.
        """
        if not self.is_path_valid(path):
            return f"Path '{path}' is outside the tool's scope."
        try:
            os.makedirs(path, exist_ok=True)
            return f"Successfully created directory '{path}'."
        except Exception as e:
            return f"An error occurred while creating directory '{path}': {e}"

    @tool
    def delete_file(self, file_path: str) -> str:
        """
        Deletes a specified file.
        :param file_path: The path of the file to delete.
        """
        if not self.is_path_valid(file_path):
            return f"Path '{file_path}' is outside the tool's scope."
        try:
            os.remove(file_path)
            return f"Successfully deleted file '{file_path}'."
        except FileNotFoundError:
            return f"Error: File '{file_path}' not found."
        except Exception as e:
            return f"An error occurred while deleting file '{file_path}': {e}"

    @tool
    def delete_directory(self, path: str, recursive: bool = False) -> str:
        """
        Deletes a specified directory.
        :param path: The path of the directory to delete.
        :param recursive: If True, deletes the directory and all its contents.
        """
        if not self.is_path_valid(path):
            return f"Path '{path}' is outside the tool's scope."
        try:
            if not os.path.exists(path):
                return f"Error: Directory '{path}' not found."
            if recursive:
                import shutil

                shutil.rmtree(path)
            else:
                os.rmdir(path)
            return f"Successfully deleted directory '{path}'."
        except OSError as e:
            return f"An error occurred while deleting directory '{path}': {e}"
        except Exception as e:
            return f"An unexpected error occurred: {e}"

    @tool
    def move_item(self, source_path: str, destination_path: str) -> str:
        """
        Moves or renames a file or directory.
        :param source_path: The current path of the file or directory.
        :param destination_path: The new path for the file or directory.
        """
        if not self.is_path_valid(source_path):
            return f"Path '{source_path}' is outside the tool's scope."
        if not self.is_path_valid(destination_path):
            return f"Path '{destination_path}' is outside the tool's scope."
        try:
            os.rename(source_path, destination_path)
            return f"Successfully moved '{source_path}' to '{destination_path}'."
        except FileNotFoundError:
            return f"Error: Source '{source_path}' not found."
        except Exception as e:
            return f"An error occurred while moving '{source_path}' to '{destination_path}': {e}"

    @tool
    def copy_file(self, source_path: str, destination_path: str) -> str:
        """
        Copies a file from source to destination.
        :param source_path: The path of the file to copy.
        :param destination_path: The destination path for the new file.
        """
        if not self.is_path_valid(source_path):
            return f"Path '{source_path}' is outside the tool's scope."
        if not self.is_path_valid(destination_path):
            return f"Path '{destination_path}' is outside the tool's scope."
        try:
            import shutil

            shutil.copy2(source_path, destination_path)
            return f"Successfully copied '{source_path}' to '{destination_path}'."
        except FileNotFoundError:
            return f"Error: Source file '{source_path}' not found."
        except Exception as e:
            return f"An error occurred while copying '{source_path}' to '{destination_path}': {e}"

    @tool
    def replace_in_file(self, file_path: str, old_string: str, new_string: str) -> str:
        """
        Replaces a string in a file, but only if it appears exactly once.
        To ensure precision, the 'old_string' should include enough context (e.g., surrounding lines)
        to uniquely identify the target location.

        :param file_path: The path of the file to modify.
        :param old_string: The exact block of text to be replaced (including context).
        :param new_string: The new block of text to insert.
        """

        if not self.is_path_valid(file_path):
            return f"Path '{file_path}' is outside the tool's scope."
        try:
            with open(file_path, encoding="utf-8") as f:
                content = f.read()

            occurrences = content.count(old_string)

            if occurrences == 0:
                return f"Error: The specified 'old_string' was not found in the file '{file_path}'. No changes were made."

            if occurrences > 1:
                return f"Error: {occurrences} occurrences found in '{file_path}'. Replacement requires a unique match."

            new_content = content.replace(old_string, new_string, 1)

            with open(file_path, "w", encoding="utf-8") as f:
                f.write(new_content)

            return f"Replacement successful in file '{file_path}'."

        except FileNotFoundError:
            return f"Error: File '{file_path}' not found."
        except Exception as e:
            return f"An error occurred: {e}"



import base64
import json

from datapizza.memory.memory import Turn
from datapizza.memory.memory_adapter import MemoryAdapter
from datapizza.type import (
    ROLE,
    FunctionCallBlock,
    FunctionCallResultBlock,
    MediaBlock,
    StructuredBlock,
    TextBlock,
)


class BedrockMemoryAdapter(MemoryAdapter):
    """Adapter for converting Memory objects to AWS Bedrock Converse API message format

    The Bedrock Converse API uses a message format similar to Anthropic's Claude API.
    """

    def _turn_to_message(self, turn: Turn) -> dict:
        """Convert a Turn to Bedrock message format"""
        content = []

        for block in turn:
            block_dict = {}

            match block:
                case TextBlock():
                    block_dict = {"text": block.content}

                case FunctionCallBlock():
                    # Bedrock uses toolUse format
                    block_dict = {
                        "toolUse": {
                            "toolUseId": block.id,
                            "name": block.name,
                            "input": block.arguments,
                        }
                    }

                case FunctionCallResultBlock():
                    # Bedrock uses toolResult format
                    block_dict = {
                        "toolResult": {
                            "toolUseId": block.id,
                            "content": [{"text": str(block.result)}],
                        }
                    }

                case StructuredBlock():
                    # Convert structured content to text
                    block_dict = {
                        "text": json.dumps(block.content)
                        if not isinstance(block.content, str)
                        else block.content,
                    }

                case MediaBlock():
                    match block.media.media_type:
                        case "image":
                            block_dict = self._process_image_block(block)
                        case "pdf":
                            block_dict = self._process_document_block(block)
                        case _:
                            raise NotImplementedError(
                                f"Unsupported media type: {block.media.media_type}"
                            )

            content.append(block_dict)

        # Bedrock expects content as a list
        return {
            "role": self._convert_role(turn.role),
            "content": content,
        }

    def _text_to_message(self, text: str, role: ROLE) -> dict:
        """Convert text and role to Bedrock message format"""
        return {
            "role": self._convert_role(role),
            "content": [{"text": text}],
        }

    def _convert_role(self, role: ROLE) -> str:
        """Convert ROLE to Bedrock role string

        Bedrock Converse API supports 'user' and 'assistant' roles
        """
        if role.value == "user":
            return "user"
        elif role.value == "assistant":
            return "assistant"
        else:
            # Default to user for system or other roles
            return "user"

    def _process_document_block(self, block: MediaBlock) -> dict:
        """Process document (PDF) blocks for Bedrock"""
        match block.media.source_type:
            case "base64":
                return {
                    "document": {
                        "format": "pdf",
                        "name": "document.pdf",
                        "source": {"bytes": base64.b64decode(block.media.source)},
                    }
                }

            case "path":
                with open(block.media.source, "rb") as f:
                    pdf_bytes = f.read()
                return {
                    "document": {
                        "format": "pdf",
                        "name": block.media.source.split("/")[-1],
                        "source": {"bytes": pdf_bytes},
                    }
                }

            case "url":
                raise NotImplementedError(
                    "Bedrock Converse API does not support document URLs directly. "
                    "Please download and provide as bytes."
                )

            case _:
                raise NotImplementedError(
                    f"Unsupported source type: {block.media.source_type}"
                )

    def _process_image_block(self, block: MediaBlock) -> dict:
        """Process image blocks for Bedrock"""
        match block.media.source_type:
            case "base64":
                return {
                    "image": {
                        "format": block.media.extension or "png",
                        "source": {"bytes": base64.b64decode(block.media.source)},
                    }
                }

            case "path":
                with open(block.media.source, "rb") as image_file:
                    image_bytes = image_file.read()
                return {
                    "image": {
                        "format": block.media.extension or "png",
                        "source": {"bytes": image_bytes},
                    }
                }

            case "url":
                raise NotImplementedError(
                    "Bedrock Converse API does not support image URLs directly. "
                    "Please download and provide as bytes."
                )

            case _:
                raise NotImplementedError(
                    f"Unsupported source type: {block.media.source_type}"
                )



import inspect
from collections.abc import Callable
from functools import wraps
from types import MethodType
from typing import Any

from .utils import (
    get_default_values,
    get_param_annotations,
    get_parameters,
    get_required_params,
)


class Tool:
    """Class that wraps a function while preserving its behavior and adding attributes."""

    def __init__(
        self,
        func: Callable | None = None,
        name: str | None = None,
        description: str | None = None,
        end: bool = False,
        properties: dict[str, dict[str, Any]] | None = None,
        required: list[str] | None = None,
        strict: bool = False,
    ):
        """
        Args:
            func (Callable | None): The function to wrap.
            name (str | None): The name of the tool.
            description (str | None): The description of the tool.
            end (bool): Whether the tool ends a chain of operations.
            properties (dict[str, dict[str, Any]] | None): The properties of the tool.
            required (list[str] | None): The required parameters of the tool.
            strict (bool): Whether the tool is strict.
        """
        self.func = func
        if not name and not func:
            raise ValueError("Must provide either name or function")

        self.name: str = name or func.__name__  # type: ignore
        self.strict: bool = strict
        self.description: str | None = description or (func.__doc__ if func else None)

        if func and (not properties or not required):
            self.required = get_required_params(inspect.signature(self.func))  # type: ignore
            param_annotations = get_param_annotations(inspect.signature(self.func))  # type: ignore
            default_values = get_default_values(inspect.signature(self.func))  # type: ignore
            self.properties = get_parameters(
                param_annotations, default_values=default_values
            )
        else:
            self.properties = properties
            self.required = required

        self.schema = self._get_function_schema()

        self.end_invoke = end
        if func:
            wraps(func)(self)

    def __call__(self, *args, **kwargs):
        if not self.func:
            raise ValueError("Function not set")

        return self.func(*args, **kwargs)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self

        bound_decorated = Tool(
            func=MethodType(self.func, instance) if self.func else None,
            name=self.name,
            description=self.description,
            end=self.end_invoke,
            properties=self.properties,
            required=self.required,
        )

        return bound_decorated

    @classmethod
    def tool_from_dict(cls, tool_dict):
        return Tool(
            func=None,
            name=tool_dict.get("name"),
            description=tool_dict.get("description"),
            end=tool_dict.get("end"),
            properties=tool_dict.get("properties"),
            required=tool_dict.get("required"),
            strict=tool_dict.get("strict"),
        )

    def _get_function_schema(self):
        return {
            "name": self.name,
            "description": self.description
            or self.func.__doc__
            or f"Function to {self.name}",
            "parameters": {
                "type": "object",
                "properties": self.properties,
                "required": self.required,
            },
        }

    def to_dict(self) -> dict:
        """Convert the tool to a dictionary for JSON serialization."""
        return {
            "name": self.name,
            "description": self.description,
            "properties": self.properties,
            "required": self.required,
            "end_invoke": self.end_invoke,
        }


def tool(func=None, name=None, description=None, end=False, strict=False):
    """
    Decorator to wrap a function in a DecoratedFunc instance.

    Can be used as @tool or @tool(name="custom_name", description="...")

    Args:
        func: The function to decorate
        **attributes: Additional attributes to attach to the function

    Returns:
        DecoratedFunc: A callable object wrapping the original function
    """

    def decorator(f):
        return Tool(f, name=name, description=description, end=end, strict=strict)

    # Handle both @tool and @tool(...)
    if func is None:
        return decorator
    return decorator(func)



import threading
import typing

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import (
    SimpleSpanProcessor,
    SpanExporter,
    SpanExportResult,
)


class InMemoryTraceExporter(SpanExporter):
    """Implementation of :class:`.SpanExporter` that stores spans in memory.
    This class can be used for testing purposes. It stores the exported spans
    in a list in memory that can be retrieved using the
    :func:`.get_finished_spans` method.
    """

    def __init__(self) -> None:
        self._finished_spans: dict[int, list[ReadableSpan]] = {}
        self._stopped = False
        self._lock = threading.Lock()

    def clear(self) -> None:
        """Clear list of collected spans."""
        with self._lock:
            self._finished_spans.clear()

    def clear_trace(self, trace_id):
        with self._lock:
            if trace_id in self._finished_spans:
                del self._finished_spans[trace_id]

    def get_finished_spans(self) -> dict[int, list[ReadableSpan]]:
        """Get list of collected spans."""
        with self._lock:
            return dict(self._finished_spans)

    def get_finished_spans_by_trace_id(self, trace_id: int) -> list[ReadableSpan]:
        with self._lock:
            return self._finished_spans.get(trace_id, [])

    def export(self, spans: typing.Sequence[ReadableSpan]) -> SpanExportResult:
        """Stores a list of spans in memory."""
        if self._stopped:
            return SpanExportResult.FAILURE
        with self._lock:
            for span in spans:
                context = span.get_span_context()
                if context:
                    if context.trace_id not in self._finished_spans:
                        self._finished_spans[context.trace_id] = []

                    self._finished_spans[context.trace_id].append(span)
        return SpanExportResult.SUCCESS

    def shutdown(self) -> None:
        """Shut downs the exporter.

        Calls to export after the exporter has been shut down will fail.
        """
        self._stopped = True

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return True


class ContextSpanProcessor(SimpleSpanProcessor):
    """Simple SpanProcessor implementation.

    SimpleSpanProcessor is an implementation of `SpanProcessor` that
    passes ended spans directly to the configured `SpanExporter`.
    """

    def __init__(self):
        super().__init__(InMemoryTraceExporter())
        self.tracing_ids = set()

    def start_trace(self, trace_id: int) -> None:
        self.tracing_ids.add(trace_id)

    def stop_trace(self, trace_id: int) -> None:
        if trace_id in self.tracing_ids:
            self.span_exporter.clear_trace(trace_id)  # type: ignore
            self.tracing_ids.remove(trace_id)

    def on_end(self, span: ReadableSpan) -> None:
        if span.get_span_context().trace_id not in self.tracing_ids:  # type: ignore
            return

        super().on_end(span)

    def get_spans_by_trace_id(self, trace_id: int) -> list[ReadableSpan]:
        return self.span_exporter.get_finished_spans_by_trace_id(trace_id)  # type: ignore
